import os

import anafora

root = '/Users/bethard/Data/thymedata'
text_roots = ['/Users/bethard/Data/THYME-corpus/train', '/Users/bethard/Data/THYME-corpus/dev']

for input_dir, output_dir, text_name, xml_names in anafora.walk_anafora_to_anafora(root):
    for text_root in text_roots:
        text_path = os.path.join(text_root, text_name)
        if os.path.exists(text_path):
            text = open(text_path).read().decode('utf-8')
            for xml_name in xml_names:
                if 'Temporal' in xml_name:
                    data = anafora.AnaforaData.from_file(os.path.join(root, input_dir, xml_name))
                    for ann in data.annotations.select_type('TIMEX3'):
                        if ann.properties['Class'] == 'SET':
                            ann_text = '...'.join(text[begin:end] for begin, end in ann.spans)
                            print(ann_text)



import anafora
import anafora.heuristic


def test_relation_to_closest():
    def get_xml():
        return anafora.ElementTree.fromstring("""
            <data>
                <annotations>
                    <entity>
                        <id>1</id>
                        <span>0,5</span>
                        <type>X</type>
                    </entity>
                    <entity>
                        <id>2</id>
                        <span>15,20</span>
                        <type>X</type>
                    </entity>
                    <entity>
                        <id>3</id>
                        <span>25,30</span>
                        <type>X</type>
                    </entity>
                    <entity>
                        <id>4</id>
                        <span>0,3</span>
                        <type>Y</type>
                    </entity>
                    <entity>
                        <id>5</id>
                        <span>21,24</span>
                        <type>Y</type>
                    </entity>
                    <entity>
                        <id>6</id>
                        <span>35,40</span>
                        <type>Y</type>
                    </entity>
                </annotations>
            </data>
            """)

    data = anafora.AnaforaData(get_xml())
    z1 = anafora.AnaforaRelation()
    z1.id = "7"
    data.annotations.append(z1)
    z1.type = "Z"
    z1.properties["source"] = data.annotations.select_id("1")
    z1.properties["target"] = data.annotations.select_id("4")
    z1.properties["foo"] = "bar"

    z2 = anafora.AnaforaRelation()
    z2.id = "8"
    data.annotations.append(z2)
    z2.type = "Z"
    z2.properties["source"] = data.annotations.select_id("2")
    z2.properties["target"] = data.annotations.select_id("5")
    z2.properties["foo"] = "bar"

    z3 = anafora.AnaforaRelation()
    z3.id = "9"
    data.annotations.append(z3)
    z3.type = "Z"
    z3.properties["source"] = data.annotations.select_id("3")
    z3.properties["target"] = data.annotations.select_id("5")
    z3.properties["foo"] = "bar"

    data = anafora.AnaforaData(get_xml())
    anafora.heuristic.add_relations_to_closest(data, "X", "Y", "Z", "source", "target", [("foo", "bar")])
    assert set(data.annotations.select_type("Z")) == {z1, z2, z3}

    # make sure it doesn't fail with 0 source and 0 target annotations
    data = anafora.AnaforaData(get_xml())
    anafora.heuristic.add_relations_to_closest(data, "A", "Y", "Z", "source", "target")
    data = anafora.AnaforaData(get_xml())
    anafora.heuristic.add_relations_to_closest(data, "X", "B", "Z", "source", "target")



import anafora
import anafora.validate

def test_schema_validate():
    schema = anafora.validate.Schema(anafora.ElementTree.fromstring("""
        <schema>
        <defaultattribute>
            <required>True</required>
        </defaultattribute>
        <definition>
            <entities>
                <entity type="X">
                        <properties>
                                <property type="A" input="choice">x,y</property>
                                <property type="B" />
                                <property type="C" instanceOf="Y,Z" />
                        </properties>
                </entity>
                <entity type="Y" />
                <entity type="Z" />
            </entities>
            <relations>
                <relation type="R">
                    <properties>
                        <property type="D" instanceOf="X" required="False" />
                        <property type="E" instanceOf="Y,Z" required="False" />
                    </properties>
                </relation>
            </relations>
        </definition>
        </schema>
        """))
    data = anafora.AnaforaData()
    entity1 = anafora.AnaforaEntity()
    entity1.id = "@1@"
    entity1.type = "X"
    entity1.properties["A"] = "x"
    data.annotations.append(entity1)
    assert schema.errors(data)
    entity1.properties["B"] = "y"
    assert schema.errors(data)
    entity1.properties["C"] = "z"
    assert schema.errors(data)
    entity2 = anafora.AnaforaEntity()
    entity2.id = "@2@"
    entity2.type = "X"
    data.annotations.append(entity2)
    entity1.properties["C"] = entity2
    assert schema.errors(data)
    entity2.type = "Y"
    assert not schema.errors(data)
    entity1.properties["A"] = "y"
    assert not schema.errors(data)
    entity1.properties["A"] = "z"
    assert schema.errors(data)
    entity1.properties["A"] = "x"
    assert not schema.errors(data)

    relation = anafora.AnaforaRelation()
    relation.id = "@3@"
    relation.type = ""
    data.annotations.append(relation)
    assert schema.errors(data)
    relation.type = "R"
    assert not schema.errors(data)
    relation.properties["D"] = entity1
    assert not schema.errors(data)
    relation.properties["E"] = entity1
    assert schema.errors(data)
    relation.properties["E"] = entity2
    assert not schema.errors(data)
    relation.properties["X"] = "Y"
    assert schema.errors(data)



import argparse
import os
import re

import anafora


def to_text(timeml_path):
    """
    :param xml.etree.ElementTree.Element timeml_path: path of the TimeML XML
    :return string: the (plain) text content of the XML
    """
    return ''.join(anafora.ElementTree.parse(timeml_path).getroot().itertext())


def to_document_creation_time(timeml_path):
    root = anafora.ElementTree.parse(timeml_path).getroot()
    dct_xpath = ".//TIMEX3[@functionInDocument='CREATION_TIME']"
    for timex3 in root.iterfind(dct_xpath):
        return timex3.attrib["value"]


def to_anafora_data(timeml_path):
    """
    :param xml.etree.ElementTree.Element timeml_path: path of the TimeML XML
    :return anafora.AnaforaData: an Anafora version of the TimeML annotations
    """
    entity_tags = {"TIMEX3", "EVENT", "SIGNAL"}
    tag_id_attrs = {
        "TIMEX3": "tid",
        "EVENT": "eid",
        "SIGNAL": "sid",
        "MAKEINSTANCE": "eiid",
        "TLINK": "lid",
        "SLINK": "lid",
        "ALINK": "lid",
    }
    ref_id_attrs = {"eventID", "signalID", "beginPoint", "endPoint", "valueFromFunction", "anchorTimeID",
                    "eventInstanceID", "timeID", "signalID", "relatedToEventInstance", "relatedToTime",
                    "subordinatedEventInstance", "tagID"}
    text = to_text(timeml_path)
    data = anafora.AnaforaData()
    root = anafora.ElementTree.parse(timeml_path).getroot()

    prefix_to_char = {'t': 'e', 'e': 'e', 's': 'e', 'ei': 'r', 'l': 'r'}
    timeml_id_to_anafora_id = {}
    count = 1
    file_base, _ = os.path.splitext(os.path.basename(timeml_path))
    for elem in root.iter():
        if elem.tag in tag_id_attrs:
            timeml_id = elem.attrib[tag_id_attrs[elem.tag]]
            [(prefix, number)] = re.findall(r'^(\D+)(\d+)$', timeml_id)
            timeml_id_to_anafora_id[timeml_id] = '{0:d}@{1}@{2}@gold'.format(count, prefix_to_char[prefix], file_base)
            count += 1

    def add_annotations_from(elem, offset=0):
        start = offset
        annotation = None
        if elem.tag in tag_id_attrs:
            annotation = anafora.AnaforaEntity() if elem.tag in entity_tags else anafora.AnaforaRelation()
            id_attr = tag_id_attrs[elem.tag]
            annotation.id = timeml_id_to_anafora_id[elem.attrib[id_attr]]
            annotation.type = elem.tag
            if isinstance(annotation, anafora.AnaforaEntity):
                annotation.spans = ((start, start),)
            for name, value in elem.attrib.items():
                if name != id_attr:
                    if name in ref_id_attrs:
                        value = timeml_id_to_anafora_id[value]
                    annotation.properties[name] = value
            data.annotations.append(annotation)

        if elem.text is not None:
            offset += len(elem.text)
        for child in elem:
            offset = add_annotations_from(child, offset)

        if annotation is not None and isinstance(annotation, anafora.AnaforaEntity):
            annotation.spans = ((start, offset),)
            if elem.text != text[start:offset]:
                raise ValueError('{0}: "{1}" != "{2}"'.format(timeml_path, elem.text, text[start:offset]))

        if elem.tail is not None:
            offset += len(elem.tail)
        return offset

    add_annotations_from(root)
    return data


def _timeml_dir_to_anafora_dir(timeml_dir, anafora_dir, schema_name="TimeML"):
    for root, _, file_names in os.walk(timeml_dir):
        if root.startswith(timeml_dir):
            sub_dir = root[len(timeml_dir):].lstrip(os.path.sep)
        else:
            sub_dir = ''

        for file_name in file_names:
            if file_name.endswith(".tml"):
                file_path = os.path.join(root, file_name)
                text = to_text(file_path)
                data = to_anafora_data(file_path)
                data.indent()

                anafora_file_name = file_name[:-4]
                anafora_file_dir = os.path.join(anafora_dir, sub_dir, anafora_file_name)
                if not os.path.exists(anafora_file_dir):
                    os.makedirs(anafora_file_dir)
                anafora_file_path = os.path.join(anafora_file_dir, anafora_file_name)

                with open(anafora_file_path, 'w') as text_file:
                    text_file.write(text)
                data.indent()
                data.to_file("{0}.{1}.gold.completed.xml".format(anafora_file_path, schema_name))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--timeml-dir", required=True)
    parser.add_argument("--anafora-dir", required=True)
    parser.add_argument("--schema-name", default="TimeML")
    args = parser.parse_args()
    _timeml_dir_to_anafora_dir(args.timeml_dir, args.anafora_dir, args.schema_name)



import pytest

import anafora
import anafora.evaluate


def test_score_data():
    reference = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1</id>
                <span>0,5</span>
                <type>X</type>
            </entity>
            <entity>
                <id>2</id>
                <span>5,10</span>
                <type>Y</type>
            </entity>
            <entity>
                <id>3</id>
                <span>15,20</span>
                <type>Y</type>
            </entity>
            <relation>
                <id>4</id>
                <type>Z</type>
                <properties>
                    <Source>1</Source>
                    <Target>2</Target>
                    <Prop1>T</Prop1>
                    <Prop2>A</Prop2>
                </properties>
            </relation>
            <relation>
                <id>5</id>
                <type>Z</type>
                <properties>
                    <Source>2</Source>
                    <Target>3</Target>
                    <Prop1>T</Prop1>
                    <Prop2>B</Prop2>
                </properties>
            </relation>
        </annotations>
    </data>
    """))
    predicted = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>6</id><!-- different -->
                <span>0,5</span>
                <type>X</type>
            </entity>
            <entity>
                <id>7</id><!-- different -->
                <span>5,10</span>
                <type>X</type><!-- different -->
            </entity>
            <entity>
                <id>8</id><!-- different -->
                <span>15,20</span>
                <type>Y</type>
            </entity>
            <relation>
                <id>9</id><!-- different -->
                <type>Z</type>
                <properties>
                    <Source>6</Source>
                    <Target>7</Target>
                    <Prop1>T</Prop1>
                    <Prop2>A</Prop2>
                </properties>
            </relation>
            <relation>
                <id>10</id><!-- different -->
                <type>Z</type>
                <properties>
                    <Source>7</Source>
                    <Target>8</Target>
                    <Prop1>F</Prop1><!-- different -->
                    <Prop2>B</Prop2>
                </properties>
            </relation>
        </annotations>
    </data>
    """))
    named_scores = anafora.evaluate.score_data(reference, predicted)
    assert set(named_scores.keys()) == {
        "*", ("*", "<span>"),
        "X", ("X", "<span>"),
        "Y", ("Y", "<span>"),
        "Z", ("Z", "<span>"), ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
    }
    scores = named_scores["X"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 2
    scores = named_scores["X", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 2
    scores = named_scores["Y"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Y", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["*"]
    assert scores.correct == 1 + 1 + 0
    assert scores.reference == 1 + 2 + 2
    assert scores.predicted == 2 + 1 + 2
    scores = named_scores["*", "<span>"]
    assert scores.correct == 1 + 1 + 2
    assert scores.reference == 1 + 2 + 2
    assert scores.predicted == 2 + 1 + 2

    named_scores = anafora.evaluate.score_data(reference, predicted, exclude=["X", "Y"])
    assert set(named_scores.keys()) == {
        "*", ("*", "<span>"),
        "Z", ("Z", "<span>"), ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
    }
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["*"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["*", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(reference, predicted, include=[("Z", "Prop1", "T")])
    assert set(named_scores.keys()) == {("Z", "Prop1", "T")}
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(reference, predicted, include=[("Z", "Prop1", "F")])
    assert set(named_scores.keys()) == {("Z", "Prop1", "F")}
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(reference, predicted, include=["Z"], exclude=[("Z", "<span>")])
    assert set(named_scores.keys()) == {
        "Z", ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
    }
    scores = named_scores["Z"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1


def test_score_data_overlap():
    # This test is identical to the one above except that the spans have been changed so that they're overlapping
    # instead of being exactly equal
    reference = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1</id>
                <span>0,5</span>
                <type>X</type>
            </entity>
            <entity>
                <id>2</id>
                <span>5,10</span>
                <type>Y</type>
            </entity>
            <entity>
                <id>3</id>
                <span>15,20</span>
                <type>Y</type>
            </entity>
            <relation>
                <id>4</id>
                <type>Z</type>
                <properties>
                    <Source>1</Source>
                    <Target>2</Target>
                    <Prop1>T</Prop1>
                    <Prop2>A</Prop2>
                </properties>
            </relation>
            <relation>
                <id>5</id>
                <type>Z</type>
                <properties>
                    <Source>2</Source>
                    <Target>3</Target>
                    <Prop1>T</Prop1>
                    <Prop2>B</Prop2>
                </properties>
            </relation>
            <relation>
                <id>6</id>
                <type>Ref</type>
                <properties>
                    <Ref>1</Ref>
                </properties>
            </relation>
            <relation>
                <id>7</id>
                <type>Ref</type>
                <properties>
                    <Ref>6</Ref>
                </properties>
            </relation>
        </annotations>
    </data>
    """))
    predicted = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>6</id><!-- different -->
                <span>0,4</span>
                <type>X</type>
            </entity>
            <entity>
                <id>7</id><!-- different -->
                <span>6,10</span>
                <type>X</type><!-- different -->
            </entity>
            <entity>
                <id>8</id><!-- different -->
                <span>19,20</span>
                <type>Y</type>
            </entity>
            <relation>
                <id>9</id><!-- different -->
                <type>Z</type>
                <properties>
                    <Source>6</Source>
                    <Target>7</Target>
                    <Prop1>T</Prop1>
                    <Prop2>A</Prop2>
                </properties>
            </relation>
            <relation>
                <id>10</id><!-- different -->
                <type>Z</type>
                <properties>
                    <Target>8</Target>
                    <Source>7</Source>
                    <Prop2>B</Prop2>
                    <Prop1>F</Prop1><!-- different -->
                </properties>
            </relation>
        </annotations>
    </data>
    """))
    named_scores = anafora.evaluate.score_data(
        reference, predicted, spans_type=anafora.evaluate._OverlappingSpans)
    assert set(named_scores.keys()) == {
        "*", ("*", "<span>"),
        "X", ("X", "<span>"),
        "Y", ("Y", "<span>"),
        "Z", ("Z", "<span>"), ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
        "Ref", ("Ref", "<span>"), ("Ref", "Ref"),
        }
    scores = named_scores["X"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 2
    scores = named_scores["X", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 2
    scores = named_scores["Y"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Y", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["*"]
    assert scores.correct == 1 + 1 + 0
    assert scores.reference == 1 + 2 + 2 + 2
    assert scores.predicted == 2 + 1 + 2
    scores = named_scores["*", "<span>"]
    assert scores.correct == 1 + 1 + 2
    assert scores.reference == 1 + 2 + 2 + 1
    assert scores.predicted == 2 + 1 + 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude=["X", "Y"],
        spans_type=anafora.evaluate._OverlappingSpans)
    assert set(named_scores.keys()) == {
        "*", ("*", "<span>"),
        "Z", ("Z", "<span>"), ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
        "Ref", ("Ref", "<span>"), ("Ref", "Ref"),
        }
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["*"]
    assert scores.correct == 1
    assert scores.reference == 2 + 2
    assert scores.predicted == 2
    scores = named_scores["*", "<span>"]
    assert scores.correct == 2
    assert scores.reference == 2 + 1
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include=[("Z", "Prop1", "T")],
        spans_type=anafora.evaluate._OverlappingSpans)
    assert set(named_scores.keys()) == {("Z", "Prop1", "T")}
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include=[("Z", "Prop1", "F")],
        spans_type=anafora.evaluate._OverlappingSpans)
    assert set(named_scores.keys()) == {("Z", "Prop1", "F")}
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include=["Z"], exclude=[("Z", "<span>")],
        spans_type=anafora.evaluate._OverlappingSpans)
    assert set(named_scores.keys()) == {
        "Z", ("Z", "Source"), ("Z", "Target"), ("Z", "Prop1"), ("Z", "Prop2"),
        ("Z", "Prop1", "T"), ("Z", "Prop1", "F"), ("Z", "Prop2", "A"), ("Z", "Prop2", "B"),
        }
    scores = named_scores["Z"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop1", "T"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop1", "F"]
    assert scores.correct == 0
    assert scores.reference == 0
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2"]
    assert scores.correct == 2
    assert scores.reference == 2
    assert scores.predicted == 2
    scores = named_scores["Z", "Prop2", "A"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "Prop2", "B"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1


def test_missing_ignored_properties():
    reference = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>4</id>
                <type>Z</type>
                <properties>
                    <A>1</A>
                    <B>2</B>
                    <C></C>
                </properties>
            </entity>
        </annotations>
    </data>
    """))
    predicted = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>4</id>
                <type>Z</type>
                <properties>
                    <B>2</B>
                    <A>1</A>
                </properties>
            </entity>
        </annotations>
    </data>
    """))
    named_scores = anafora.evaluate.score_data(reference, predicted)
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1

    # make sure no exceptions are thrown
    anafora.evaluate._print_document_scores([("temp", named_scores)])

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude=[("Z", "C")])
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1


def test_temporal_closure_scores():

    def annotation(source, target, value):
        return (source, target), None, (None, value)

    reference = {
        annotation("A", "B", "BEFORE"),
        annotation("B", "C", "IS_INCLUDED"),
        annotation("D", "C", "INCLUDES"),
        annotation("E", "D", "CONTAINS"),
        annotation("F", "E", "AFTER"),
        annotation("G", "H", "BEGINS-ON"),
        annotation("I", "G", "BEFORE"),
        annotation("J", "K", "IBEFORE"),
        annotation("K", "L", "BEGUN_BY"),
        annotation("L", "K", "BEGINS"), # duplicate
        # inferred:
        # A before B
        # A before F
        # B before F
        # C includes B
        # C before F
        # D includes B
        # D includes C
        # D before F
        # E includes B
        # E includes C
        # E includes D
        # E before F
        # G simultaneous-start H
        # I before G
        # I before H
    }
    predicted = {
        annotation("A", "B", "BEFORE"),   # (+)
        annotation("A", "B", "BEFORE"),   # duplicate
        annotation("B", "A", "AFTER"),    # duplicate
        annotation("B", "E", "CONTAINS"), # (-)
        annotation("B", "E", "INCLUDES"), # duplicate
        annotation("B", "F", "BEFORE"),   # (+)
        annotation("F", "D", "AFTER"),    # (+)
        annotation("H", "I", "AFTER"),    # (+)
        annotation("J", "L", "IBEFORE"),  # (+)
        annotation("K", "L", "BEGUN_BY"), # (+)
        # inferred:
        # (+) A before B
        # ( ) A before E
        # ( ) A before F
        # ( ) B includes E
        # ( ) B before F
        # ( ) D before F
        # (+) E before F
        # ( ) G after I
        # ( ) J i-before L
        # (+) K begun-by L
        # (+) J i-before K
    }
    scores = anafora.evaluate.TemporalClosureScores()
    scores.add(reference, predicted)
    assert scores.precision() == 6.0 / 7.0
    assert scores.recall() ==  4.0 / 9.0


def test_temporal_closure_data():
    reference = anafora.AnaforaData(anafora.ElementTree.fromstring("""
        <data>
            <annotations>
                <entity>
                    <id>67@e@ID006_clinic_016@gold</id>
                    <span>2220,2231</span>
                    <type>TIMEX3</type>
                    <parentsType>TemporalEntities</parentsType>
                    <properties>
                        <Class>DATE</Class>
                    </properties>
                </entity>
                <entity>
                    <id>20@e@ID006_clinic_016@gold</id>
                    <span>2211,2219</span>
                    <type>EVENT</type>
                    <parentsType>TemporalEntities</parentsType>
                    <properties>
                        <DocTimeRel>BEFORE</DocTimeRel>
                        <Type>N/A</Type>
                        <Degree>N/A</Degree>
                        <Polarity>POS</Polarity>
                        <ContextualModality>ACTUAL</ContextualModality>
                        <ContextualAspect>N/A</ContextualAspect>
                        <Permanence>UNDETERMINED</Permanence>
                    </properties>
                </entity>
                <entity>
                    <id>22@e@ID006_clinic_016@gold</id>
                    <span>2273,2279</span>
                    <type>EVENT</type>
                    <parentsType>TemporalEntities</parentsType>
                    <properties>
                        <DocTimeRel>BEFORE</DocTimeRel>
                        <Type>N/A</Type>
                        <Degree>N/A</Degree>
                        <Polarity>POS</Polarity>
                        <ContextualModality>ACTUAL</ContextualModality>
                        <ContextualAspect>N/A</ContextualAspect>
                        <Permanence>UNDETERMINED</Permanence>
                    </properties>
                </entity>
                <relation>
                    <id>16@r@ID006_clinic_016@gold</id>
                    <type>TLINK</type>
                    <parentsType>TemporalRelations</parentsType>
                    <properties>
                        <Source>67@e@ID006_clinic_016@gold</Source>
                        <Type>CONTAINS</Type>
                        <Target>20@e@ID006_clinic_016@gold</Target>
                    </properties>
                </relation>
                <relation>
                    <id>36@r@ID006_clinic_016@gold</id>
                    <type>TLINK</type>
                    <parentsType>TemporalRelations</parentsType>
                    <properties>
                        <Source>20@e@ID006_clinic_016@gold</Source>
                        <Type>CONTAINS</Type>
                        <Target>22@e@ID006_clinic_016@gold</Target>
                    </properties>
                </relation>
            </annotations>
        </data>
        """))
    predicted = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>56@regex</id>
                <type>EVENT</type>
                <span>2211,2219</span>
                <properties>
                    <Polarity>POS</Polarity>
                    <Degree>N/A</Degree>
                    <Permanence>UNDETERMINED</Permanence>
                    <ContextualAspect>N/A</ContextualAspect>
                    <DocTimeRel>OVERLAP</DocTimeRel>
                    <Type>N/A</Type>
                    <ContextualModality>ACTUAL</ContextualModality>
                </properties>
            </entity>
            <entity>
                <id>57@regex</id>
                <type>TIMEX3</type>
                <span>2220,2231</span>
                <properties>
                    <Class>DATE</Class>
                </properties>
            </entity>
            <entity>
                <id>58@regex</id>
                <type>EVENT</type>
                <span>2273,2279</span>
                <properties>
                    <Polarity>POS</Polarity>
                    <Degree>N/A</Degree>
                    <Permanence>UNDETERMINED</Permanence>
                    <ContextualAspect>N/A</ContextualAspect>
                    <DocTimeRel>BEFORE/OVERLAP</DocTimeRel>
                    <Type>EVIDENTIAL</Type>
                    <ContextualModality>ACTUAL</ContextualModality>
                </properties>
            </entity>
            <relation>
                <id>57@regex@TLINK@58@regex</id>
                <type>TLINK</type>
                <properties>
                    <Source>57@regex</Source>
                    <Target>58@regex</Target>
                    <Type>CONTAINS</Type>
                    <Extra>42</Extra>
                </properties>
            </relation>
            <relation>
                <id>55@regex@TLINK@58@regex</id>
                <type>TLINK</type>
                <properties>
                    <Source>56@regex</Source>
                    <Target>58@regex</Target>
                    <Type>CONTAINS</Type>
                </properties>
            </relation>
        </annotations>
    </data>
    """))
    # reference: T(2220,2231) -> E(2211,2219) -> E(2273,2279)
    # predicted: T(2220,2231) -> E(2211,2219); E(2273,2279) -> E(2211,2219)

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include={"TLINK"})
    scores = named_scores["TLINK"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include={("TLINK", "Type", "CONTAINS")})
    scores = named_scores["TLINK", "Type", "CONTAINS"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include={("TLINK", "Type", "CONTAINS")},
        scores_type=anafora.evaluate.TemporalClosureScores)
    scores = named_scores["TLINK", "Type", "CONTAINS"]
    assert scores.precision_correct == 2
    assert scores.recall_correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"EVENT"}, include={"TLINK"})
    scores = named_scores["TLINK"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"TIMEX3"}, include={"TLINK"})
    scores = named_scores["TLINK"]
    assert scores.correct == 0
    assert scores.reference == 2
    assert scores.predicted == 2

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"EVENT", "TIMEX3"}, include={"TLINK"})
    scores = named_scores["TLINK"]
    assert scores.correct == 1
    assert scores.reference == 2
    assert scores.predicted == 2

    with pytest.raises(RuntimeError) as exc_info:
        anafora.evaluate.score_data(
            reference, predicted, include={"EVENT"},
            scores_type=anafora.evaluate.TemporalClosureScores)
    assert "binary spans" in str(exc_info.value)

    with pytest.raises(RuntimeError) as exc_info:
        anafora.evaluate.score_data(
            reference, predicted, include={"TLINK"},
            scores_type=anafora.evaluate.TemporalClosureScores)
    assert "single property" in str(exc_info.value)


def test_delete_excluded():
    reference = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1@e</id>
                <type>Z</type>
                <span>1, 3</span>
                <properties>
                    <A>2@e</A>
                </properties>
            </entity>
            <entity>
                <id>2@e</id>
                <type>Y</type>
                <span>4, 6</span>
                <properties>
                    <B>3@e</B>
                </properties>
            </entity>
            <entity>
                <id>3@e</id>
                <type>X</type>
                <span>7, 9</span>
            </entity>
            <entity>
                <id>4@e</id>
                <type>W</type>
                <span>20, 30</span>
                <properties>
                    <B>3@e</B>
                </properties>
            </entity>
        </annotations>
    </data>
    """))
    predicted = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>4@e</id>
                <type>Z</type>
                <span>1, 3</span>
                <properties>
                    <A>5@e</A>
                </properties>
            </entity>
            <entity>
                <id>5@e</id>
                <type>Y</type>
                <span>4, 6</span>
                <properties>
                    <B>6@e</B>
                </properties>
            </entity>
            <entity>
                <id>6@e</id>
                <type>X</type>
                <span>10, 15</span>
            </entity>
            <entity>
                <id>7@e</id>
                <type>W</type>
                <span>20, 30</span>
                <properties>
                    <B></B>
                </properties>
            </entity>
        </annotations>
    </data>
    """))
    named_scores = anafora.evaluate.score_data(reference, predicted)
    scores = named_scores["X"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Y"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["W"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"X"})
    scores = named_scores["Y"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["W"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"Y"})
    scores = named_scores["X"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, exclude={"Z"})
    scores = named_scores["X"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Y"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1

    named_scores = anafora.evaluate.score_data(
        reference, predicted, include={("*", "<span>")})
    scores = named_scores["X", "<span>"]
    assert scores.correct == 0
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Y", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["Z", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1
    scores = named_scores["W", "<span>"]
    assert scores.correct == 1
    assert scores.reference == 1
    assert scores.predicted == 1



import anafora
import anafora.select


def test_select_all():
    select = anafora.select.Select()

    assert select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert select('C')


def test_select_type():
    select = anafora.select.Select(include={'A', 'B'})

    assert select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(exclude=[('A',), ('B',)])

    assert not select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(include={'A'}, exclude={'B'})

    assert select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')


def test_select_prop_name():
    select = anafora.select.Select(include=[('A',), ('B', 'R')])

    assert select('A')
    assert not select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(exclude={'A', ('B', 'R')})

    assert not select('A')
    assert select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(include={'B'}, exclude={('B', 'P')})

    assert not select('A')
    assert not select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')


def select_prop_value():
    select = anafora.select.Select(include={'A', ('B', 'R', 'V')})

    assert select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(exclude={'A', ('B', 'R', 'V')})

    assert not select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert not select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(include={'C', ('C', 'R', 'W')})

    assert not select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert select('C')


def test_select_star():
    select = anafora.select.Select(include={'*'})

    assert select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(exclude={'*'})

    assert not select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(include={('*', 'R')})

    assert not select('A')
    assert not select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(include=[('C',)], exclude=[('C', '*')])

    assert not select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert not select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(include=['C', 'B'], exclude=[('C', '*'), ('B', '*', '*')])

    assert not select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert not select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(exclude={('C', 'R', '*')})

    assert select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert select('C')

    select = anafora.select.Select(include={('*', '*', '*')})

    assert not select('A')
    assert not select('B', 'P')
    assert not select('B', 'R')
    assert select('B', 'R', 'V')
    assert select('C', 'R', 'V')
    assert select('C', 'R', 'W')
    assert not select('C')

    select = anafora.select.Select(exclude={('*', '*', '*')})

    assert select('A')
    assert select('B', 'P')
    assert select('B', 'R')
    assert not select('B', 'R', 'V')
    assert not select('C', 'R', 'V')
    assert not select('C', 'R', 'W')
    assert select('C')



import sys
import textwrap
from optparse import Values
from typing import List

from pip._internal.cli.base_command import Command
from pip._internal.cli.status_codes import SUCCESS
from pip._internal.utils.misc import get_prog

BASE_COMPLETION = """
# pip {shell} completion start{script}# pip {shell} completion end
"""

COMPLETION_SCRIPTS = {
    "bash": """
        _pip_completion()
        {{
            COMPREPLY=( $( COMP_WORDS="${{COMP_WORDS[*]}}" \\
                           COMP_CWORD=$COMP_CWORD \\
                           PIP_AUTO_COMPLETE=1 $1 2>/dev/null ) )
        }}
        complete -o default -F _pip_completion {prog}
    """,
    "zsh": """
        function _pip_completion {{
          local words cword
          read -Ac words
          read -cn cword
          reply=( $( COMP_WORDS="$words[*]" \\
                     COMP_CWORD=$(( cword-1 )) \\
                     PIP_AUTO_COMPLETE=1 $words[1] 2>/dev/null ))
        }}
        compctl -K _pip_completion {prog}
    """,
    "fish": """
        function __fish_complete_pip
            set -lx COMP_WORDS (commandline -o) ""
            set -lx COMP_CWORD ( \\
                math (contains -i -- (commandline -t) $COMP_WORDS)-1 \\
            )
            set -lx PIP_AUTO_COMPLETE 1
            string split \\  -- (eval $COMP_WORDS[1])
        end
        complete -fa "(__fish_complete_pip)" -c {prog}
    """,
}


class CompletionCommand(Command):
    """A helper command to be used for command completion."""

    ignore_require_venv = True

    def add_options(self) -> None:
        self.cmd_opts.add_option(
            "--bash",
            "-b",
            action="store_const",
            const="bash",
            dest="shell",
            help="Emit completion code for bash",
        )
        self.cmd_opts.add_option(
            "--zsh",
            "-z",
            action="store_const",
            const="zsh",
            dest="shell",
            help="Emit completion code for zsh",
        )
        self.cmd_opts.add_option(
            "--fish",
            "-f",
            action="store_const",
            const="fish",
            dest="shell",
            help="Emit completion code for fish",
        )

        self.parser.insert_option_group(0, self.cmd_opts)

    def run(self, options: Values, args: List[str]) -> int:
        """Prints the completion code of the given shell"""
        shells = COMPLETION_SCRIPTS.keys()
        shell_options = ["--" + shell for shell in sorted(shells)]
        if options.shell in shells:
            script = textwrap.dedent(
                COMPLETION_SCRIPTS.get(options.shell, "").format(prog=get_prog())
            )
            print(BASE_COMPLETION.format(script=script, shell=options.shell))
            return SUCCESS
        else:
            sys.stderr.write(
                "ERROR: You must pass {}\n".format(" or ".join(shell_options))
            )
            return SUCCESS



import sys
import os
import re
import importlib
import warnings
import contextlib


is_pypy = '__pypy__' in sys.builtin_module_names


warnings.filterwarnings('ignore',
                        r'.+ distutils\b.+ deprecated',
                        DeprecationWarning)


def warn_distutils_present():
    if 'distutils' not in sys.modules:
        return
    if is_pypy and sys.version_info < (3, 7):
        # PyPy for 3.6 unconditionally imports distutils, so bypass the warning
        # https://foss.heptapod.net/pypy/pypy/-/blob/be829135bc0d758997b3566062999ee8b23872b4/lib-python/3/site.py#L250
        return
    warnings.warn(
        "Distutils was imported before Setuptools, but importing Setuptools "
        "also replaces the `distutils` module in `sys.modules`. This may lead "
        "to undesirable behaviors or errors. To avoid these issues, avoid "
        "using distutils directly, ensure that setuptools is installed in the "
        "traditional way (e.g. not an editable install), and/or make sure "
        "that setuptools is always imported before distutils.")


def clear_distutils():
    if 'distutils' not in sys.modules:
        return
    warnings.warn("Setuptools is replacing distutils.")
    mods = [name for name in sys.modules if re.match(r'distutils\b', name)]
    for name in mods:
        del sys.modules[name]


def enabled():
    """
    Allow selection of distutils by environment variable.
    """
    which = os.environ.get('SETUPTOOLS_USE_DISTUTILS', 'local')
    return which == 'local'


def ensure_local_distutils():
    clear_distutils()

    # With the DistutilsMetaFinder in place,
    # perform an import to cause distutils to be
    # loaded from setuptools._distutils. Ref #2906.
    with shim():
        importlib.import_module('distutils')

    # check that submodules load as expected
    core = importlib.import_module('distutils.core')
    assert '_distutils' in core.__file__, core.__file__


def do_override():
    """
    Ensure that the local copy of distutils is preferred over stdlib.

    See https://github.com/pypa/setuptools/issues/417#issuecomment-392298401
    for more motivation.
    """
    if enabled():
        warn_distutils_present()
        ensure_local_distutils()


class DistutilsMetaFinder:
    def find_spec(self, fullname, path, target=None):
        if path is not None:
            return

        method_name = 'spec_for_{fullname}'.format(**locals())
        method = getattr(self, method_name, lambda: None)
        return method()

    def spec_for_distutils(self):
        import importlib.abc
        import importlib.util

        try:
            mod = importlib.import_module('setuptools._distutils')
        except Exception:
            # There are a couple of cases where setuptools._distutils
            # may not be present:
            # - An older Setuptools without a local distutils is
            #   taking precedence. Ref #2957.
            # - Path manipulation during sitecustomize removes
            #   setuptools from the path but only after the hook
            #   has been loaded. Ref #2980.
            # In either case, fall back to stdlib behavior.
            return

        class DistutilsLoader(importlib.abc.Loader):

            def create_module(self, spec):
                return mod

            def exec_module(self, module):
                pass

        return importlib.util.spec_from_loader('distutils', DistutilsLoader())

    def spec_for_pip(self):
        """
        Ensure stdlib distutils when running under pip.
        See pypa/pip#8761 for rationale.
        """
        if self.pip_imported_during_build():
            return
        clear_distutils()
        self.spec_for_distutils = lambda: None

    @classmethod
    def pip_imported_during_build(cls):
        """
        Detect if pip is being imported in a build script. Ref #2355.
        """
        import traceback
        return any(
            cls.frame_file_is_setup(frame)
            for frame, line in traceback.walk_stack(None)
        )

    @staticmethod
    def frame_file_is_setup(frame):
        """
        Return True if the indicated frame suggests a setup.py file.
        """
        # some frames may not have __file__ (#2940)
        return frame.f_globals.get('__file__', '').endswith('setup.py')


DISTUTILS_FINDER = DistutilsMetaFinder()


def add_shim():
    DISTUTILS_FINDER in sys.meta_path or insert_shim()


@contextlib.contextmanager
def shim():
    insert_shim()
    try:
        yield
    finally:
        remove_shim()


def insert_shim():
    sys.meta_path.insert(0, DISTUTILS_FINDER)


def remove_shim():
    try:
        sys.meta_path.remove(DISTUTILS_FINDER)
    except ValueError:
        pass



import pytest

import anafora


def test_empty():
    data = anafora.AnaforaData(anafora.ElementTree.fromstring('<data/>'))
    assert list(data.annotations) == []

    data = anafora.AnaforaData(anafora.ElementTree.fromstring('<data><annotations></annotations></data>'))
    assert list(data.annotations) == []


def test_duplicate_id():
    with pytest.raises(ValueError):
        anafora.AnaforaData(anafora.ElementTree.fromstring('''
        <data>
            <annotations>
                <entity><id>1</id></entity>
                <entity><id>1</id></entity>
            </annotations>
        </data>'''))

    data = anafora.AnaforaData()
    entity1 = anafora.AnaforaEntity()
    entity1.id = "1"
    entity2 = anafora.AnaforaEntity()
    entity2.id = "1"
    data.annotations.append(entity1)
    with pytest.raises(ValueError):
        data.annotations.append(entity2)


def test_add_entity():
    data = anafora.AnaforaData()
    assert str(data) == '<data />'
    entity = anafora.AnaforaEntity()
    with pytest.raises(ValueError) as exception_info:
        data.annotations.append(entity)
    assert "id" in str(exception_info.value)
    assert str(data) == '<data />'
    entity.id = "1"
    data.annotations.append(entity)
    assert str(data) == '<data><annotations><entity><id>1</id></entity></annotations></data>'
    entity.type = "X"
    entity.parents_type = "Y"
    entity.properties["name1"] = "value1"
    assert str(data) == ('<data><annotations><entity>' +
                         '<id>1</id>' +
                         '<type>X</type>' +
                         '<parentsType>Y</parentsType>' +
                         '<properties><name1>value1</name1></properties>' +
                         '</entity></annotations></data>')
    del entity.properties["name1"]
    assert str(data) == ('<data><annotations><entity>' +
                         '<id>1</id>' +
                         '<type>X</type>' +
                         '<parentsType>Y</parentsType>' +
                         '</entity></annotations></data>')
    with pytest.raises(ValueError):
        del entity.properties["name2"]


def test_add_reference():
    data = anafora.AnaforaData()
    entity1 = anafora.AnaforaEntity()
    entity1.id = "@1@"
    entity2 = anafora.AnaforaEntity()
    entity2.id = "@2@"
    with pytest.raises(ValueError) as exception_info:
        entity2.properties["link"] = entity1
    assert "<annotations" in str(exception_info.value)
    data.annotations.append(entity1)
    with pytest.raises(ValueError):
        entity2.properties["link"] = entity1
    assert "<annotations" in str(exception_info.value)
    data.annotations.append(entity2)
    entity2.properties["link"] = entity1
    assert str(data) == ('<data><annotations>' +
                         '<entity><id>@1@</id></entity>' +
                         '<entity><id>@2@</id><properties><link>@1@</link></properties></entity>' +
                         '</annotations></data>')


def test_remove():
    data = anafora.AnaforaData()
    assert str(data) == '<data />'
    entity1 = anafora.AnaforaEntity()
    entity1.id = "@1@"
    data.annotations.append(entity1)
    entity2 = anafora.AnaforaEntity()
    entity2.id = "@2@"
    entity2.properties["name"] = "value"
    data.annotations.append(entity2)
    assert list(data.annotations) == [entity1, entity2]
    assert str(data) == ('<data><annotations>' +
                         '<entity><id>@1@</id></entity>' +
                         '<entity><id>@2@</id><properties><name>value</name></properties></entity>' +
                         '</annotations></data>')
    data.annotations.remove(entity1)
    assert list(data.annotations) == [entity2]
    assert str(data) == ('<data><annotations>' +
                         '<entity><id>@2@</id><properties><name>value</name></properties></entity>' +
                         '</annotations></data>')
    data.annotations.remove(entity2)
    assert list(data.annotations) == []
    assert str(data) == '<data><annotations /></data>'


def test_spans():
    data = anafora.AnaforaData(anafora.ElementTree.fromstring('''
        <data>
            <annotations>
                <relation>
                    <id>1</id>
                    <type>R1</type>
                    <properties>
                        <relation>2</relation>
                    </properties>
                </relation>
                <relation>
                    <id>2</id>
                    <type>R2</type>
                    <properties>
                        <entity>3</entity>
                    </properties>
                </relation>
                <entity>
                    <id>3</id>
                    <type>E1</type>
                    <span>5,7</span>
                </entity>
            </annotations>
        </data>'''))
    assert data.annotations.select_id("1").spans == ((((5, 7),),),)
    assert data.annotations.select_id("2").spans == (((5, 7),),)
    assert data.annotations.select_id("3").spans == ((5, 7),)


def test_recursive_entity():
    data = anafora.AnaforaData()
    entity = anafora.AnaforaEntity()
    entity.id = "@1@"
    data.annotations.append(entity)
    entity.properties["self"] = entity
    assert entity.is_self_referential()
    assert data.annotations.find_self_referential().id == entity.id

    data = anafora.AnaforaData()
    a = anafora.AnaforaEntity()
    a.id = "A"
    data.annotations.append(a)
    b = anafora.AnaforaEntity()
    b.id = "B"
    data.annotations.append(b)
    c = anafora.AnaforaEntity()
    c.id = "C"
    data.annotations.append(c)
    d = anafora.AnaforaEntity()
    d.id = "D"
    data.annotations.append(d)
    b.properties["x"] = a
    c.properties["y"] = a
    d.properties["1"] = b
    d.properties["2"] = c
    assert not d.is_self_referential()



def test_sort():
    data = anafora.AnaforaData(anafora.ElementTree.fromstring('''
        <data>
            <annotations>
                <entity>
                    <id>1</id>
                    <type>E</type>
                    <span>5,7</span>
                </entity>
                <entity>
                    <id>2</id>
                    <type>E</type>
                    <span>3,4</span>
                </entity>
            </annotations>
        </data>'''))
    assert [a.id for a in sorted(data.annotations)] == ['2', '1']




import argparse
import os

import anafora


class Select(object):
    def __init__(self, include=None, exclude=None):
        self.include = None
        if include is not None:
            self.include = set()
            for item in include:
                self.include.add(item if isinstance(item, tuple) else (item,))
        self.exclude = None
        if exclude is not None:
            self.exclude = set()
            for item in exclude:
                self.exclude.add(item if isinstance(item, tuple) else (item,))

    @staticmethod
    def _expand(*args):
        args = [a for a in args if a is not None]
        result = set()
        if len(args) == 1:
            result.add((args[0],))
            result.add(('*',))
        else:
            for rest in Select._expand(*args[1:]):
                result.add((args[0],) + rest)
                result.add(('*',) + rest)
        return result

    def is_included(self, type_name, prop_name=None, prop_value=None):
        if self.include is not None:
            if not (Select._expand(type_name) & self.include):
                if not (Select._expand(type_name, prop_name) & self.include):
                    if not (Select._expand(type_name, prop_name, prop_value) & self.include):
                        return False
        return True

    def is_excluded(self, type_name, prop_name=None, prop_value=None):
        if self.exclude is not None:
            if Select._expand(type_name) & self.exclude:
                return True
            if Select._expand(type_name, prop_name) & self.exclude:
                return True
            if Select._expand(type_name, prop_name, prop_value) & self.exclude:
                return True
        return False

    def __call__(self, type_name, prop_name=None, prop_value=None):
        return self.is_included(type_name, prop_name, prop_value) and \
               not self.is_excluded(type_name, prop_name, prop_value)


def _main(input_dir, output_dir, xml_name_regex="[.]xml$", include=None, exclude=None):
    select = Select(include, exclude)

    for sub_dir, text_name, xml_names in anafora.walk(input_dir, xml_name_regex):
        for xml_name in xml_names:

            # reads in the data from the input file
            xml_path = os.path.join(input_dir, sub_dir, xml_name)
            data = anafora.AnaforaData.from_file(xml_path)

            # find annotations and properties to remove
            annotations_to_remove = []
            annotation_properties_to_remove = []
            for annotation in data.annotations:

                # remove the annotation if its type has not been selected
                if not select(annotation.type):
                    annotations_to_remove.append(annotation)
                else:
                    for name, value in annotation.properties.items():

                        # remove the property if its name or value has not been selected
                        if not select(annotation.type, name, value):
                            annotation_properties_to_remove.append((annotation, name))

            # if we're overwriting, save a backup of the original
            if annotations_to_remove or annotation_properties_to_remove:
                data.to_file(xml_path + ".bak")

            # do the actual removal of annotations here
            for annotation in annotations_to_remove:
                data.annotations.remove(annotation)
            for annotation, name in annotation_properties_to_remove:
                del annotation.properties[name]

            # writes out the modified data to the output file
            output_sub_dir = os.path.join(output_dir or input_dir, sub_dir)
            if not os.path.exists(output_sub_dir):
                os.makedirs(output_sub_dir)
            output_path = os.path.join(output_sub_dir, xml_name)
            data.to_file(output_path)


if __name__ == "__main__":
    def split_tuple_on_colons(string):
        return tuple(string.split(":"))

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="input_dir",
                        help="The root of a set of Anafora annotation XML directories.")
    parser.add_argument("-o", "--output", metavar="DIR", dest="output_dir",
                        help="The directory where the cleaned versions of the Anafora annotation XML files should be " +
                             "written. The directory structure will mirror the input directory structure.")
    parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                        help="A regular expression for matching XML files in the input subdirectories " +
                             "(default: %(default)r)")
    parser.add_argument("--include", metavar="EXPR", nargs="+", type=split_tuple_on_colons,
                        help="An expression identifying types of annotations to be included in the data. " +
                             "The expression takes the form type[:property[:value]. For example, TLINK would only " +
                             "include TLINK annotations (and TLINK properties and property values) in the " +
                             "evaluation, while TLINK:Type:CONTAINS would only include TLINK annotations with a Type " +
                             "property that has the value CONTAINS.")
    parser.add_argument("--exclude", metavar="EXPR", nargs="+", type=split_tuple_on_colons,
                        help="An expression identifying types of annotations to be excluded from the data. " +
                             "The expression takes the form type[:property[:value] (see --include).")

    args = parser.parse_args()
    _main(**vars(args))


import sys
import subprocess


def __optim_args_from_interpreter_flags():
    """Return a list of command-line arguments reproducing the current
    optimization settings in sys.flags."""
    args = []
    value = sys.flags.optimize
    if value > 0:
        args.append("-" + "O" * value)
    return args


_optim_args_from_interpreter_flags = getattr(
    subprocess,
    "_optim_args_from_interpreter_flags",
    __optim_args_from_interpreter_flags,
)



import anafora
import anafora.regex


def test_regex_annotator():
    annotator = anafora.regex.RegexAnnotator({
        'aa+': ('A', {}),
        'a': ('A', {'X': '2'}),
        'bb': ('B', {'Y': '1'})
    })
    text = "bb aaa"
    data = anafora.AnaforaData()
    annotator.annotate(text, data)

    assert len(list(data.annotations)) == 2
    [b_annotation, a_annotation] = data.annotations
    assert b_annotation.type == "B"
    assert b_annotation.spans == ((0, 2),)
    assert dict(b_annotation.properties.items()) == {'Y': '1'}
    assert a_annotation.type == "A"
    assert a_annotation.spans == ((3, 6),)
    assert dict(a_annotation.properties.items()) == {}


def test_preannotated():
    annotator = anafora.regex.RegexAnnotator({
        'aa+': ('A', {'X': '2'}),
        'a': ('A', {}),
        'bb': ('B', {'Y': '1'})
    }, {
        'C': {'Z': '3'}
    })
    text = "bb aaa"
    data = anafora.AnaforaData()
    bb = anafora.AnaforaEntity()
    bb.id = "1@preannotated"
    bb.type = "B"
    bb.spans = ((0, 2),)
    data.annotations.append(bb)
    aaa = anafora.AnaforaEntity()
    aaa.id = "2@preannotated"
    aaa.type = "C"
    aaa.spans = ((3, 6),)
    data.annotations.append(aaa)
    annotator.annotate(text, data)

    assert len(list(data.annotations)) == 3
    [b_annotation, c_annotation, a_annotation] = data.annotations
    assert b_annotation.type == "B"
    assert b_annotation.spans == ((0, 2),)
    assert dict(b_annotation.properties.items()) == {'Y': '1'}
    assert c_annotation.type == "C"
    assert c_annotation.spans == ((3, 6),)
    assert dict(c_annotation.properties.items()) == {'Z': '3'}
    assert a_annotation.type == "A"
    assert a_annotation.spans == ((3, 6),)
    assert dict(a_annotation.properties.items()) == {'X': '2'}


def test_many_groups():
    regex_predictions = {}
    for i in range(1, 1000):
        regex_predictions['a' * i] = ('A' * i, {})
    annotator = anafora.regex.RegexAnnotator(regex_predictions)
    text = "aaaaaaaaaa"
    data = anafora.AnaforaData()
    annotator.annotate(text, data)

    assert len(list(data.annotations)) == 1
    [annotation] = data.annotations
    assert annotation.type == "AAAAAAAAAA"
    assert annotation.spans == ((0, 10),)
    assert dict(annotation.properties.items()) == {}


def test_file_roundtrip(tmpdir):
    annotator_path = str(tmpdir.join("temp.annotator"))
    annotator = anafora.regex.RegexAnnotator({
        'the year': ('DATE', {}),
        'John': ('PERSON', {'type': 'NAME', 'gender': 'MALE'}),
        '.1.2.\d+;': ('OTHER', {})
    }, {
        'PERSON': {'type': 'NAME', 'gender': 'FEMALE'}
    })
    annotator.to_file(annotator_path)
    assert anafora.regex.RegexAnnotator.from_file(annotator_path) == annotator


def test_simple_file(tmpdir):
    path = tmpdir.join("temp.annotator")
    path.write("""\
aaa aaa\tA
b\tB\t{"x": "y"}
\\dc\\s+x\tC
""")
    annotator = anafora.regex.RegexAnnotator({
        'aaa aaa': ('A', {}),
        'b': ('B', {'x': 'y'}),
        r'\dc\s+x': ('C', {})
    })
    assert anafora.regex.RegexAnnotator.from_file(str(path)) == annotator


def test_train():
    text1 = "aaa bb ccccc dddd"
    data1 = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1</id>
                <type>AA</type>
                <span>0,6</span><!-- "aaa bb" -->
                <properties>
                    <a>A</a>
                </properties>
            </entity>
            <entity>
                <id>2</id>
                <type>AA</type>
                <span>7,12</span><!-- "ccccc" -->
                <properties>
                    <c>B</c>
                </properties>
            </entity>
            <entity>
                <id>3</id>
                <type>EMPTY</type>
                <span>0,0</span>
            </entity>
        </annotations>
    </data>
    """))
    text2 = "ccccc dddd ccccc dddd ccccc."
    data2 = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1</id>
                <type>CC</type>
                <span>0,5</span><!-- "ccccc" -->
                <properties>
                    <c>B</c>
                </properties>
            </entity>
            <entity>
                <id>2</id>
                <type>CC</type>
                <span>11,16</span><!-- "ccccc" -->
                <properties>
                    <c>C</c>
                </properties>
            </entity>
            <entity>
                <id>3</id>
                <type>CC</type>
                <span>22,27</span><!-- "ccccc" -->
                <properties>
                    <c>C</c>
                    <d>D</d>
                </properties>
            </entity>
            <entity>
                <id>4</id>
                <type>PERIOD</type>
                <span>27,28</span><!-- "." -->
            </entity>
        </annotations>
    </data>
    """))

    annotator = anafora.regex.RegexAnnotator({
        r'\baaa\s+bb\b': ('AA', {"a": "A"}),
        r'\bccccc\b': ('CC', {"c": "C", "d": "D"}),
        r'\b\.': ('PERIOD', {}),
    }, {
        'AA': {'a': 'A', 'c': 'B'},
        'CC': {'c': 'C', 'd': 'D'},
    })
    assert anafora.regex.RegexAnnotator.train([(text1, data1), (text2, data2)]) == annotator


def test_filter_by_precision():
    annotator = anafora.regex.RegexAnnotator({
        r'the': ("THE", {}),
        r'\bthe\b': ("THE", {}),
        r'yer\b': ("ER", {}),
        r'er\b': ("ER", {})
    })
    text = "the theater near the record player"
    data = anafora.AnaforaData(anafora.ElementTree.fromstring("""
    <data>
        <annotations>
            <entity>
                <id>1</id>
                <type>THE</type>
                <span>0,3</span><!-- "the" -->
            </entity>
            <entity>
                <id>2</id>
                <type>THE</type>
                <span>17,20</span><!-- "the" -->
            </entity>
            <entity>
                <id>3</id>
                <type>ER</type>
                <span>9,11</span><!-- "er" -->
            </entity>
            <entity>
                <id>4</id>
                <type>ER</type>
                <span>32,34</span><!-- "." -->
            </entity>
        </annotations>
    </data>
    """))
    annotator.prune_by_precision(0.6, [(text, data)])
    assert annotator == anafora.regex.RegexAnnotator({
        r'the': ("THE", {}),
        r'\bthe\b': ("THE", {}),
        r'er\b': ("ER", {})
    })
    annotator.prune_by_precision(1.0, [(text, data)])
    assert annotator == anafora.regex.RegexAnnotator({
        r'\bthe\b': ("THE", {}),
        r'er\b': ("ER", {})
    })



import argparse
import ast
import itertools
import json
import logging
import os
import re
from typing import Any, Text, Mapping
import xml.etree.cElementTree as ET


def _iter_schemas(anafora_setting: ET.Element, schema_name=None):
    for an_schema_elem in anafora_setting.findall("./schemas/schema"):
        an_schema_name = an_schema_elem.attrib["name"]

        # the <file> may be listed directly
        if schema_name is None or schema_name == an_schema_name:
            an_schema_path = an_schema_elem.findtext("file")
            if an_schema_path:
                yield an_schema_name, an_schema_path

        # or there may be <mode>s, each with their own <file>
        for an_mode_elem in an_schema_elem.iter('mode'):
            an_mode_name = an_mode_elem.attrib["name"]
            an_schema_path = an_mode_elem.find("file").text
            an_schema_mode_name = f"{an_schema_name}-{an_mode_name}"
            if schema_name is None or schema_name == an_schema_mode_name:
                yield an_schema_mode_name, an_schema_path


def _iter_anafora_annotation_paths(
        anafora_path: Text,
        schema: Text,
        annotator: Text,
        status: Text,
        project: Text):
    an_filename_matcher = re.compile(
        f".*[.]{schema}[.]{annotator}[.]{status}[.]xml")
    for dirpath, dirnames, filenames in os.walk(anafora_path):
        if project is None or project in dirpath:
            for filename in filenames:
                if an_filename_matcher.match(filename):
                    yield os.path.join(anafora_path, dirpath, filename)


def anafora_to_labelstudio(
        anafora_path: Text,
        labelstudio_path: Text,
        annotator: Text,
        status: Text,
        project: Text,
        schema: Text):

    # iterate over the <schema> elements
    an_setting_tree = ET.parse(os.path.join(anafora_path, ".setting.xml"))
    for an_schema_name, an_schema_path in _iter_schemas(
            anafora_setting=an_setting_tree.getroot(),
            schema_name=schema):
        an_schema_path = os.path.join(anafora_path, ".schema", an_schema_path)
        if os.path.exists(an_schema_path):
            ls_tree, ls_property_types = anafora_schema_to_labelstudio_schema(
                anafora_tree=ET.parse(an_schema_path),
            )
            ET.indent(ls_tree, space="  ", level=0)
            ls_tree.write(f"{labelstudio_path}.{an_schema_name}.schema.xml")

            ls_data = []
            for an_annotations_path in _iter_anafora_annotation_paths(
                    anafora_path=anafora_path,
                    schema=an_schema_name,
                    annotator=annotator,
                    status=status,
                    project=project):
                logging.info(f"Processing {an_annotations_path}")
                text_path, _, _, _, _ = an_annotations_path.rsplit('.', 4)
                if not os.path.exists(text_path):
                    logging.warning(f"Skipping file because text file is "
                                    f"missing:\n{an_annotations_path}")
                    continue
                with open(text_path) as text_file:
                    text = text_file.read()
                ls_data.append(anafora_annotations_to_labelstudio_annotations(
                    anafora_tree=ET.parse(an_annotations_path),
                    text=text,
                    source=os.path.basename(text_path),
                    labelstudio_property_types=ls_property_types,
                ))

            ls_output_path = f"{labelstudio_path}.{an_schema_name}.data.json"
            with open(ls_output_path, 'w') as labelstudio_file:
                json.dump(ls_data, labelstudio_file, indent=4)


def anafora_schema_to_labelstudio_schema(
        anafora_tree: ET.ElementTree) -> tuple[ET.ElementTree, Mapping[Text, Text]]:
    # the overall view
    ls_view_elem = ET.Element('View', dict(style="display: flex;"))

    # the view of the label choices
    ls_labels_view_elem = ET.SubElement(ls_view_elem, 'View', dict(
        style="flex: 20%"))
    ls_labels_elem = ET.SubElement(ls_labels_view_elem, 'Labels', dict(
        name="type", toName="text", showInline="false"))
    ls_relations_elem = ET.SubElement(ls_labels_view_elem, 'Relations')
    relation_types = set()

    # the view of the text
    ls_text_view_elem = ET.SubElement(ls_view_elem, 'View', dict(
        style="flex: 60%"))
    ET.SubElement(ls_text_view_elem, 'Text', dict(name="text", value="$text"))

    # the view of the attribute choices
    ls_attrib_view_elem = ET.SubElement(ls_view_elem, 'View', dict(
        style="flex: 20%"))

    an_root = anafora_tree.getroot()
    default_attributes = dict(required=False)
    for an_defaults_elem in an_root.iter("defaultattribute"):
        for elem in an_defaults_elem:
            default_attributes[elem.tag] = ast.literal_eval(elem.text)

    def get(attrib, name):
        return attrib.get(name, default_attributes[name])

    ls_property_types = {}
    for an_entities_elem in an_root.iter('entities'):
        entities_type = an_entities_elem.attrib["type"]
        ET.SubElement(ls_labels_elem, 'Header', dict(value=f"{entities_type}:"))
        for an_entity_elem in an_entities_elem.iter('entity'):
            entity_type = an_entity_elem.attrib["type"]
            ls_label_attrib = dict(value=entity_type)
            hotkey = an_entity_elem.attrib.get("hotkey")
            if hotkey is not None:
                ls_label_attrib["hotkey"] = hotkey
            color = an_entity_elem.attrib.get("color")
            if color is not None:
                ls_label_attrib["background"] = f"#{color}"
            ET.SubElement(ls_labels_elem, 'Label', ls_label_attrib)
            for an_property_elem in an_entity_elem.iter('property'):
                property_type = an_property_elem.attrib["type"]
                property_input = an_property_elem.attrib["input"]
                property_required = get(an_property_elem.attrib, "required")
                ls_property_name = f"{entity_type}-{property_type}"
                ls_prop_attrib = dict(
                    visibleWhen="region-selected",
                    whenTagName="type",
                    whenLabelValue=entity_type)

                if property_input == "text":
                    ls_property_types[ls_property_name] = "textarea"
                    ls_prop_elem = ET.SubElement(
                        ls_attrib_view_elem, 'View', ls_prop_attrib)
                    ET.SubElement(ls_prop_elem, 'Header', dict(
                        value=f"{property_type}:"))
                    ls_text_area_attrib = dict(
                        name=ls_property_name,
                        toName="text",
                        perRegion="true")
                    if property_required:
                        ls_text_area_attrib["required"] = "true"
                    ET.SubElement(ls_prop_elem, 'TextArea', ls_text_area_attrib)

                elif property_input == "choice":
                    ls_property_types[ls_property_name] = "choices"
                    ls_prop_elem = ET.SubElement(
                        ls_attrib_view_elem, 'View', ls_prop_attrib)
                    ls_choices_attrib = dict(
                        name=ls_property_name,
                        toName="text",
                        perRegion="true")
                    if property_required:
                        ls_choices_attrib["required"] = "true"
                    ls_choices_elem = ET.SubElement(
                        ls_prop_elem, 'Choices', ls_choices_attrib)
                    ET.SubElement(ls_choices_elem, 'Header', dict(
                        value=f"{property_type}:"))
                    for choice in an_property_elem.text.split(','):
                        if choice:
                            ET.SubElement(ls_choices_elem, 'Choice', dict(
                                value=choice))

                elif property_input == "list":
                    ls_property_types[ls_property_name] = 'relation'
                    if property_type not in relation_types:
                        ET.SubElement(ls_relations_elem, 'Relation', dict(
                            value=property_type))
                        relation_types.add(property_type)

                else:
                    raise ValueError(f'unexpected input_type: {property_input}')

    for an_relations_elem in an_root.iter('relations'):
        relations_type = an_relations_elem.attrib["type"]
        for an_relation_elem in an_relations_elem.iter('relation'):
            relation_type = an_relation_elem.attrib["type"]
            argument_types = []
            relation_subtypes = []
            for an_property_elem in an_relation_elem.iter('property'):
                property_type = an_property_elem.attrib["type"]
                property_input = an_property_elem.attrib["input"]
                ls_property_name = f"{relation_type}-{property_type}"
                if property_input == "choice":
                    relation_subtypes.append([
                        f"{property_type}={value}"
                        for value in an_property_elem.text.split(',')])
                    ls_property_types[ls_property_name] = 'choices'
                elif property_input == "list":
                    ls_property_types[ls_property_name] = 'relation'
                    argument_types.append(property_type)
                else:
                    raise ValueError(f'unexpected input_type: {property_input}')

            source_type = argument_types[0]
            for argument_type in argument_types[1:]:
                if not relation_subtypes:
                    relation_strings = [f"{relation_type}:{source_type}:"
                                        f"{argument_type}"]
                else:
                    relation_strings = [
                        f"{relation_type}:{':'.join(subtypes)}:{source_type}:"
                        f"{argument_type}"
                        for subtypes in itertools.product(*relation_subtypes)
                    ]
                for relation_string in relation_strings:
                    ET.SubElement(ls_relations_elem, 'Relation', dict(
                        value=relation_string))
    ls_tree = ET.ElementTree(ls_view_elem)
    return ls_tree, ls_property_types


def anafora_annotations_to_labelstudio_annotations(
        anafora_tree: ET.ElementTree,
        text: Text,
        source: Text,
        labelstudio_property_types: Mapping[Text, Text]) -> Mapping[Text, Any]:
    ls_types = labelstudio_property_types
    ls_type_to_value = {"textarea": "text", "choices": "choices"}
    ls_results = []
    ls_meta_info = {"source": source}

    an_root = anafora_tree.getroot()
    an_info_elem = an_root.find('info')
    for an_elem in an_info_elem:
        ls_meta_info[an_elem.tag] = an_elem.text

    an_annotations_elem = an_root.find("annotations")
    for an_elem in an_annotations_elem:
        an_id = an_elem.find("id").text
        an_type = an_elem.find("type").text
        an_parents_type = an_elem.find("parentsType").text

        if an_elem.tag == "entity":
            an_spans = [
                tuple(int(offset) for offset in tuple(span_text.split(",")))
                for span_text in an_elem.find("span").text.split(";")
            ]
            for i, (start, end) in enumerate(an_spans):
                an_revised_id = an_id if i == 0 else f"{an_id}-{i}"
                ls_results.append({
                    "value": {
                        "start": start,
                        "end": end,
                        "labels": [an_type],
                    },
                    "id": an_revised_id,
                    "from_name": "type",
                    "to_name": "text",
                    "type": "labels"
                })
                for an_prop_elem in an_elem.find('properties'):
                    an_prop_name = an_prop_elem.tag
                    an_prop_value = an_prop_elem.text
                    if an_prop_value:
                        ls_property_name = f"{an_type}-{an_prop_name}"
                        ls_property_type = ls_types[ls_property_name]
                        if ls_property_type == 'relation':
                            ls_results.append({
                                "from_id": an_revised_id,
                                "to_id": an_prop_value,
                                "type": "relation",
                                "labels": [an_prop_name],
                            })
                        else:
                            ls_results.append({
                                "value": {
                                    "start": start,
                                    "end": end,
                                    ls_type_to_value[ls_property_type]: [
                                        an_prop_value
                                    ],
                                },
                                "id": an_revised_id,
                                "from_name": ls_property_name,
                                "to_name": "text",
                                "type": ls_property_type
                            })

        elif an_elem.tag == "relation":
            choices_elems = []
            relations_elems = []
            for an_prop_elem in an_elem.find('properties'):
                ls_property_type = labelstudio_property_types[f"{an_type}-{an_prop_elem.tag}"]
                if ls_property_type == "choices":
                    choices_elems.append(an_prop_elem)
                elif ls_property_type == "relation":
                    relations_elems.append(an_prop_elem)
                else:
                    raise ValueError(f'unexpected property type: {ls_property_type}')

            relation_string = an_type
            for an_prop_elem in choices_elems:
                relation_string += f":{an_prop_elem.tag}={an_prop_elem.text}"

            ls_source_id = relations_elems[0].text
            ls_source_type = relations_elems[0].tag
            for an_prop_elem in relations_elems[1:]:
                ls_target_id = an_prop_elem.text
                ls_target_type = an_prop_elem.tag
                ls_results.append({
                    "from_id": ls_source_id,
                    "to_id": ls_target_id,
                    "type": "relation",
                    "from_name": "",
                    "labels": [
                        f"{relation_string}:{ls_source_type}:{ls_target_type}"
                    ]
                })
        else:
            raise ValueError(f'unexpected element type: {an_elem.tag}')

    return {
        "data": {
            "text": text,
            "meta_info": ls_meta_info
        },
        "annotations": [{"result": ls_results}]
    }


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""%(prog)s converts Anafora schema XML and Anafora
        annotations XML into Label Studio schema XML and Label Studio
        annotations JSON.""")
    parser.add_argument("anafora_path",
                        help="""Path to the root directory of an Anafora
                        installation, including the .setting.xml file, the
                        .schema directory, and the directories of XML data
                        files.""")
    parser.add_argument("labelstudio_path",
                        help="""Prefix of the Label Studio XML schema and JSON
                        annotation files to be written out.""")
    parser.add_argument("--annotator", default="gold", metavar="NAME",
                        help="""Selects which annotator's files should be
                        converted (default: %(default)s).""")
    parser.add_argument("--status", default="completed", metavar="TYPE",
                        help="""Selects which status of annotations should be
                        converted (default: %(default)s).""")
    parser.add_argument("--project", metavar="NAME",
                        help="""Selects which annotation projects should be
                        converted (default: all)""")
    parser.add_argument("--schema", metavar="NAME",
                        help="""Selects which annotation schemas should be
                        converted (default: all)""")
    anafora_to_labelstudio(**vars(parser.parse_args()))



__author__ = 'bethard'

import argparse
import collections
import copy
import functools
import glob
import logging
import os
import re

import anafora
import anafora.select


class Scores(object):

    def __init__(self):
        self.reference = 0
        self.predicted = 0
        self.correct = 0

    def add(self, reference, predicted):
        """
        :param set reference: the reference annotations
        :param set predicted: the predicted annotations
        """
        self.reference += len(reference)
        self.predicted += len(predicted)
        self.correct += len(reference & predicted)

    def update(self, other):
        """
        :param Scores other: scores to merge into this one
        """
        self.reference += other.reference
        self.predicted += other.predicted
        self.correct += other.correct

    def precision(self):
        """
        :return float: the fraction of predicted annotations that were correct
        """
        return 1.0 if self.predicted == 0 else self.correct / float(self.predicted)

    def recall(self):
        """
        :return float: the fraction of reference annotations that were found
        """
        return 1.0 if self.reference == 0 else self.correct / float(self.reference)

    def f1(self):
        """
        :return float: the harmonic mean of precision and recall
        """
        p = self.precision()
        r = self.recall()
        return 0.0 if p + r == 0.0 else 2 * p * r / (p + r)

    def __repr__(self):
        return "{0}(reference={1}, predicted={2}, correct={3})".format(
            self.__class__.__name__, self.reference, self.predicted, self.correct
        )


class DebuggingScores(Scores):
    def __init__(self):
        Scores.__init__(self)
        self.errors = []

    def add(self, reference, predicted):
        """
        :param set reference: the reference annotations
        :param set predicted: the predicted annotations
        """
        Scores.add(self, reference, predicted)
        errors = []
        for item in reference - predicted:
            errors.append((item, "not in predicted"))
        for item in predicted - reference:
            errors.append((item, "not in reference"))
        errors.sort()
        self.errors.extend(errors)

    def update(self, other):
        """
        :param DebuggingScores other: scores to merge into this one
        """
        Scores.update(self, other)
        self.errors.extend(other.errors)


class TemporalClosureScores(object):
    def __init__(self):
        self.reference = 0
        self.predicted = 0
        self.precision_correct = 0
        self.recall_correct = 0

    @property
    def correct(self):
        return self.precision_correct, self.recall_correct

    def add(self, reference, predicted):
        """
        :param set reference: the reference annotations
        :param set predicted: the predicted annotations
        """
        reference = {a for a in reference if self._is_valid(a)}
        predicted = {a for a in predicted if self._is_valid(a)}
        reference = self._remove_duplicate_relations(reference)
        predicted = self._remove_duplicate_relations(predicted)
        self.reference += len(reference)
        self.predicted += len(predicted)
        self.precision_correct += len(self._closure(reference) & predicted)
        self.recall_correct += len(reference & self._closure(predicted))

    def update(self, other):
        """
        :param TemporalClosureScores other: scores to merge into this one
        """
        self.reference += other.reference
        self.predicted += other.predicted
        self.precision_correct += other.precision_correct
        self.recall_correct += other.recall_correct

    def precision(self):
        """
        :return float: the fraction of predicted annotations that were correct (or inferable)
        """
        return 1.0 if self.predicted == 0 else self.precision_correct / float(self.predicted)

    def recall(self):
        """
        :return float: the fraction of reference annotations that were found (or inferable)
        """
        return 1.0 if self.reference == 0 else self.recall_correct / float(self.reference)

    def f1(self):
        """
        :return float: the harmonic mean of precision and recall
        """
        p = self.precision()
        r = self.recall()
        return 0.0 if p + r == 0.0 else 2 * p * r / (p + r)

    def __repr__(self):
        return "{0}(reference={1}, predicted={2}, precision_correct={3}, recall_correct={4})".format(
            self.__class__.__name__, self.reference, self.predicted, self.precision_correct, self.recall_correct
        )

    def _is_valid(self, annotation):

        # temporal closure only makes sense with binary relations
        try:
            (start, end), _, prop = annotation
        except (TypeError, ValueError):
            msg = "temporal closure requires binary spans, found {0}"
            raise RuntimeError(msg.format(annotation))

        try:
            name, value = prop
        except (TypeError, ValueError):
            msg = "temporal closure requires a single property, found {0}"
            raise RuntimeError(msg.format(annotation))

        # temporal closure only works on a defined set of temporal relations
        if value not in self._interval_to_point:
            logging.warning("invalid relation for temporal closure {0}".format(annotation))
            return False

        # otherwise, temporal closure should work
        return True

    def _remove_duplicate_relations(self, annotations):
        seen_point_relations = set()
        result_annotations = set()
        for annotation in annotations:

            # only include this annotation if no previous annotation expanded to the same point relations
            point_relations = frozenset(self._to_point_relations(annotation))
            if point_relations not in seen_point_relations:
                seen_point_relations.add(point_relations)
                result_annotations.add(annotation)

        # return the filtered annotations
        return result_annotations

    def _to_point_relations(self, annotation):
        start = self._start
        end = self._end

        # converts an interval relation to point relations
        point_relations = set()
        intervals, _, (_, value) = annotation
        interval1, interval2 = intervals

        # the start of an interval is always before its end
        point_relations.add(((interval1, start), "<", (interval1, end)))
        point_relations.add(((interval2, start), "<", (interval2, end)))

        # use the interval-to-point lookup table to add the necessary point relations
        for index1, side1, relation, index2, side2 in self._interval_to_point[value]:
            point1 = (intervals[index1], side1)
            point2 = (intervals[index2], side2)
            point_relations.add((point1, relation, point2))

            # for reflexive point relations, add them in the other direction too
            if relation == "=":
                point_relations.add((point2, relation, point1))

        # return the collected relations
        return point_relations

    def _to_interval_relations(self, point_relations, annotations):
        # map intervals to names
        interval_names = collections.defaultdict(set)
        for spans, type_name, (prop_name, _) in annotations:
            for span in spans:
                interval_names[span].add((type_name, prop_name))

        # find all pairs of intervals that have some point relation between them (and whose names match)
        pair_names = {}
        for ((interval1, _), _, (interval2, _)) in point_relations:
            names = interval_names[interval1] & interval_names[interval2]
            if names:
                pair_names[(interval1, interval2)] = names
                pair_names[(interval2, interval1)] = names

        # for each interval pair, see if it satisfies the point-wise requirements for any interval relations
        interval_relations = set()
        for pair in pair_names:
            names = pair_names[pair]
            for relation, requirements in self._interval_to_point.items():
                if all(((pair[i1], s1), r, (pair[i2], s2)) in point_relations
                       for i1, s1, r, i2, s2 in requirements):
                    for type_name, prop_name in names:
                        interval_relations.add((pair, type_name, (prop_name, relation)))

        # return the collected relations
        return interval_relations

    def _closure(self, annotations):

        # convert interval relations to point relations
        new_relations = {r for a in annotations for r in self._to_point_relations(a)}

        # repeatedly apply point transitivity rules until no new relations can be inferred
        point_relations = set()
        point_relations_index = collections.defaultdict(set)
        while new_relations:

            # update the result and the index with any new relations found on the last iteration
            point_relations.update(new_relations)
            for point_relation in new_relations:
                point_relations_index[point_relation[0]].add(point_relation)

            # infer any new transitive relations, e.g., if A < B and B < C then A < C
            new_relations = set()
            for point1, relation12, point2 in point_relations:
                for _, relation23, point3 in point_relations_index[point2]:
                    relation13 = self._point_transitions[relation12][relation23]
                    new_relation = (point1, relation13, point3)
                    if new_relation not in point_relations:
                        new_relations.add(new_relation)

        # convert the point relations back to interval relations
        return self._to_interval_relations(point_relations, annotations)

    # constants representing the start point and end point of an interval
    _start = 0
    _end = 1

    # mapping from interval relation names to point relations
    # for example, BEFORE means that the first interval's end is before the second interval's start
    _interval_to_point = {
        "BEFORE": [(0, _end, "<", 1, _start)],
        "AFTER": [(1, _end, "<", 0, _start)],
        "IBEFORE": [(0, _end, "=", 1, _start)],
        "IAFTER": [(0, _start, "=", 1, _end)],
        "CONTAINS": [(0, _start, "<", 1, _start), (1, _end, "<", 0, _end)],
        "INCLUDES": [(0, _start, "<", 1, _start), (1, _end, "<", 0, _end)],
        "IS_INCLUDED": [(1, _start, "<", 0, _start), (0, _end, "<", 1, _end)],
        "BEGINS-ON": [(0, _start, "=", 1, _start)],
        "ENDS-ON": [(0, _end, "=", 1, _end)],
        "BEGINS": [(0, _start, "=", 1, _start), (0, _end, "<", 1, _end)],
        "BEGUN_BY": [(0, _start, "=", 1, _start), (1, _end, "<", 0, _end)],
        "ENDS":  [(1, _start, "<", 0, _start), (0, _end, "=", 1, _end)],
        "ENDED_BY":  [(0, _start, "<", 1, _start), (0, _end, "=", 1, _end)],
        "SIMULTANEOUS": [(0, _start, "=", 1, _start), (0, _end, "=", 1, _end)],
        "IDENTITY": [(0, _start, "=", 1, _start), (0, _end, "=", 1, _end)],
        "DURING": [(0, _start, "=", 1, _start), (0, _end, "=", 1, _end)],
        "DURING_INV": [(0, _start, "=", 1, _start), (0, _end, "=", 1, _end)],
        "OVERLAP": [(0, _start, "<", 1, _end), (1, _start, "<", 0, _end)],
    }

    # transitivity table for point relations
    _point_transitions = {
        "<": {"<": "<", "=": "<"},
        "=": {"<": "<", "=": "="},
    }


@functools.total_ordering
class _OverlappingSpans(object):
    def __init__(self, spans):
        self.spans = spans

    def __iter__(self):
        return iter(self.spans)

    def __eq__(self, other):
        for self_start, self_end in self.spans:
            for other_start, other_end in other.spans:
                if self_start < other_end and other_start < self_end:
                    return True
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return 0

    def __lt__(self, other):
        return self.spans < other.spans

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, self.spans)


class ToSet(object):
    def __init__(self,
                 select,
                 spans_type=None,
                 type_name="*",
                 prop_name="*",
                 prop_value="*"):
        self.select = select
        self.spans_type = spans_type
        self.type_name = type_name
        self.prop_name = prop_name
        self.prop_value = prop_value

    def accept(self, annotation):
        if self.select(annotation.type, self.prop_name, self.prop_value) or \
                self.select(annotation.type, "<span>"):
            if self.type_name == "*" or annotation.type == self.type_name:
                if self.prop_name == "*" or self.prop_value == "*":
                    return True
                if self.prop_name is not None:
                    if self.prop_name in annotation.properties:
                        if annotation.properties[self.prop_name] == self.prop_value:
                            return True
        return False

    def key(self, annotation):
        if not isinstance(annotation, anafora.AnaforaAnnotation):
            return annotation
        spans = self._spans(annotation)
        props = None
        if self.prop_name == "*":
            props = []
            for name in sorted(annotation.properties):
                value = annotation.properties[name]
                if value is None:
                    continue
                if annotation.type == self.type_name:
                    if not self.select(annotation.type, name, value):
                        continue
                if isinstance(value, anafora.AnaforaAnnotation):
                    if self.select.is_excluded(value.type):
                        continue
                props.append((name, self.key(value)))
            props = tuple(props)
        elif self.prop_name is not None and annotation.type == self.type_name:
            if self.select(annotation.type, self.prop_name, self.prop_value):
                if self.prop_name in annotation.properties:
                    value = self.key(annotation.properties[self.prop_name])
                    props = self.prop_name, value
        return spans, annotation.type, props

    def _spans(self, annotation):
        if isinstance(annotation, anafora.AnaforaEntity):
            spans = annotation.spans
            if self.spans_type is not None:
                spans = self.spans_type(spans)
        elif isinstance(annotation, anafora.AnaforaRelation):
            spans = tuple(
                self._spans(annotation.properties[prop_name])
                for prop_name in sorted(annotation.properties)
                if isinstance(annotation.properties[prop_name], anafora.AnaforaAnnotation))
            if len(spans) == 1:
                spans = spans[0]
        else:
            raise ValueError("unknown annotation type: {0}".format(annotation))
        return spans

    def __call__(self, iterable):
        return {self.key(x) for x in iterable if self.accept(x)}


def score_data(reference_data, predicted_data, include=None, exclude=None,
               scores_type=Scores, spans_type=None):
    """
    :param AnaforaData reference_data: reference ("gold standard") Anafora data
    :param AnaforaData predicted_data: predicted (system-generated) Anafora data
    :param set include: types of annotations to include (others will be excluded); may be type names,
        (type-name, property-name) tuples, (type-name, property-name, property-value) tuples
    :param set exclude: types of annotations to exclude; may be type names, (type-name, property-name) tuples,
        (type-name, property-name, property-value) tuples
    :param type scores_type: type for calculating matches between predictions and reference
    :param type spans_type: wrapper object to apply to annotation spans
    :return dict: mapping from (annotation type[, property name[, property value]]) to Scores object
    """

    # returns true if this type:property:value is accepted by includes= and excludes=
    select = anafora.select.Select(include, exclude)

    # get reference and predicted annotations
    reference_annotations = reference_data.annotations
    predicted_annotations = [] if predicted_data is None else predicted_data.annotations

    # determines available views by examining all the annotations
    span = "<span>"
    views = {}
    if select("*"):
        views["*"] = ToSet(select=select,
                           spans_type=spans_type)
    if select("*", span):
        views["*", span] = ToSet(select=select,
                                 spans_type=spans_type,
                                 prop_name=None)
    for annotations in [reference_annotations, predicted_annotations]:
        for ann in annotations:
            if ann.type not in views:
                if select(ann.type):
                    views[ann.type] = ToSet(select=select,
                                            spans_type=spans_type,
                                            type_name=ann.type)
            if (ann.type, span) not in views:
                if select(ann.type, span):
                    views[ann.type, span] = ToSet(select=select,
                                                  spans_type=spans_type,
                                                  type_name=ann.type,
                                                  prop_name=None)
            for prop_name, prop_value in ann.properties.items():
                if (ann.type, prop_name) not in views:
                    if select(ann.type, prop_name):
                        views[ann.type, prop_name] = ToSet(
                            select=select,
                            spans_type=spans_type,
                            type_name=ann.type,
                            prop_name=prop_name)
                if not isinstance(prop_value, anafora.AnaforaAnnotation):
                    if (ann.type, prop_name, prop_value) not in views:
                        if select(ann.type, prop_name, prop_value):
                            if prop_value is not None:
                                views[ann.type, prop_name, prop_value] = ToSet(
                                    select=select,
                                    spans_type=spans_type,
                                    type_name=ann.type,
                                    prop_name=prop_name,
                                    prop_value=prop_value)

    # fill a mapping from a name (type, type:property or type:property:value) to the corresponding scores
    result = collections.defaultdict(lambda: scores_type())
    for view_name in sorted(views, key=lambda x: x if isinstance(x, tuple) else (x,)):
        to_set = views[view_name]
        set1 = to_set(reference_annotations)
        set2 = to_set(predicted_annotations)
        result[view_name].add(set1, set2)

    # return the collected scores
    return result


def _load(xml_path):
    """
    Tries to load data from an Anafora XML file, issuing errors on failure.

    :param xml_path: the path to an Anafora XML file
    :return AnaforaData: the data loaded from the XML, or None if there was a failure
    """
    if not os.path.exists(xml_path):
        logging.warn("%s: no such file", xml_path)
        return None
    try:
        data = anafora.AnaforaData.from_file(xml_path)
    except anafora.ElementTree.ParseError:
        logging.warn("%s: ignoring invalid XML", xml_path)
        return None
    else:
        return data


def score_dirs(reference_dir, predicted_dir, xml_name_regex="[.]xml$", text_dir=None,
               include=None, exclude=None, scores_type=Scores, spans_type=None):
    """
    :param string reference_dir: directory containing reference ("gold standard") Anafora XML directories
    :param string predicted_dir: directory containing predicted (system-generated) Anafora XML directories
    :param xml_name_regex: regular expression matching the files to be compared
    :param string text_dir: directory containing the raw texts corresponding to the Anafora XML
        (if None, texts are assumed to be in the reference dir)
    :param set include: types of annotations to include (others will be excluded); may be type names,
        (type-name, property-name) tuples, (type-name, property-name, property-value) tuples
    :param set exclude: types of annotations to exclude; may be type names, (type-name, property-name) tuples,
        (type-name, property-name, property-value) tuples
    :param type scores_type: type for calculating matches between predictions and reference
    :param type spans_type: wrapper object to apply to annotation spans
    :return iter: an iterator of (file-name, name-to-scores) where name-to-scores is a mapping from
        (annotation type[, property name[, property value]]) to a Scores object
    """

    # walks through the reference Anafora XML directories, scoring each and adding those to the overall scores
    for sub_dir, text_name, reference_xml_names in anafora.walk(reference_dir, xml_name_regex):

        # load the reference data from its Anafora XML
        try:
            [reference_xml_name] = reference_xml_names
        except ValueError:
            logging.warn("expected one reference file for %s, found %s", text_name, reference_xml_names)
            if not reference_xml_names:
                continue
            reference_xml_name = reference_xml_names[0]
        reference_xml_path = os.path.join(reference_dir, sub_dir, reference_xml_name)
        reference_data = _load(reference_xml_path)

        # check for self-references in the annotations, which cause equality and hashing to fail
        self_reference = reference_data.annotations.find_self_referential()
        if self_reference is not None:
            msg = "skipping reference file %s with self-referential annotation %s"
            logging.warn(msg, reference_xml_path, self_reference.id)
            continue

        # find and load the corresponding predicted data from its Anafora XML
        predicted_xml_glob = os.path.join(predicted_dir, sub_dir, text_name + "*.xml")
        predicted_xml_paths = [f for f in glob.glob(predicted_xml_glob) if re.search(xml_name_regex, f) is not None]
        try:
            [predicted_xml_path] = predicted_xml_paths
            predicted_data = _load(predicted_xml_path)
        except ValueError:
            logging.warn("expected one predicted file at %s, found %s", predicted_xml_glob, predicted_xml_paths)
            if not predicted_xml_paths:
                predicted_xml_path = None
                predicted_data = anafora.AnaforaData()
            else:
                predicted_xml_path = predicted_xml_paths[0]
                predicted_data = _load(predicted_xml_path)

        # check for self-references in the annotations, which cause equality and hashing to fail
        self_reference = predicted_data.annotations.find_self_referential()
        if self_reference is not None:
            msg = "skipping predicted file %s with self-referential annotation %s"
            logging.warn(msg, predicted_xml_path, self_reference.id)
            predicted_data = anafora.AnaforaData()

        # determine the path for the raw text source file
        if text_dir is None:
            text_path = os.path.join(reference_dir, sub_dir, text_name)
        else:
            text_path = os.path.join(text_dir, text_name)

        # if no raw text was found, then asking for the text of an annotation is an error
        if not os.path.exists(text_path) or not os.path.isfile(text_path):
            def _span_text(_):
                raise RuntimeError("no text file found at {0}".format(text_path))

        # otherwise, the text of an annotation can be extracted based on its spans
        else:
            with open(text_path) as text_file:
                text = text_file.read()

            def _flatten(items):
                if isinstance(items, tuple) and isinstance(items[0], int):
                    yield items
                else:
                    for item in items:
                        for flattened_items in _flatten(item):
                            yield flattened_items

            def _span_text(spans):
                return "...".join(text[start:end] for start, end in _flatten(spans))

        # score this data and update the overall scores
        named_scores = score_data(reference_data, predicted_data, include, exclude,
                                  scores_type=scores_type, spans_type=spans_type)
        for name, scores in named_scores.items():

            # if there were some predictions, and if we're using scores that keep track of errors, log the errors
            if predicted_xml_paths:
                for annotation, message in getattr(scores, "errors", []):
                    spans, _, _ = annotation
                    logging.debug('%s: %s: "%s" %s"', text_name, message, _span_text(spans), annotation)

        # generate the file name and the resulting scores
        yield text_name, named_scores


def score_annotators(anafora_dir, xml_name_regex, include=None, exclude=None,
                     scores_type=Scores, spans_type=None):
    """
    :param anafora_dir: directory containing Anafora XML directories
    :param xml_name_regex: regular expression matching the annotator files to be compared
    :param include: types of annotations to include (others will be excluded); may be type names,
        (type-name, property-name) tuples, (type-name, property-name, property-value) tuples
    :param set exclude: types of annotations to exclude; may be type names, (type-name, property-name) tuples,
        (type-name, property-name, property-value) tuples
    :param type scores_type: type for calculating matches between predictions and reference
    :param type spans_type: wrapper object to apply to annotation spans
    :return iter: an iterator of (file-name, name-to-scores) where name-to-scores is a mapping from
        (annotation type[, property name[, property value]]) to a Scores object
    """

    # pattern for extracting the annotator name from the Anafora XML file name
    annotator_name_regex = "([^.]*)[.][^.]*[.]xml$"

    # function for getting a canonical prefix corresponding to a pair of annotators
    def make_prefix(annotators):
        return "{0}-vs-{1}".format(*sorted(annotators))

    # walks through the Anafora XML directories, scoring each and adding those to the overall scores
    for sub_dir, text_name, xml_names in anafora.walk(anafora_dir, xml_name_regex):

        # load the data from each Anafora XML file
        annotator_data = []
        for xml_name in xml_names:

            # ignore in-progress annotations and automatic pre-annotations
            if '.inprogress.' in xml_name or '.preannotation.' in xml_name:
                continue

            # ignore empty files
            xml_path = os.path.join(anafora_dir, sub_dir, xml_name)
            if os.stat(xml_path).st_size == 0:
                continue

            # load the data and add it to the list
            data = _load(xml_path)
            annotator_name = re.search(annotator_name_regex, xml_name).group(1)
            annotator_data.append((annotator_name, data))

        # at least 2 annotators are needed for annotator agreement
        if len(annotator_data) < 2:
            logging.warn("%s: found fewer than 2 annotators: %s", text_name, xml_names)
            continue

        # pair each annotator with each other annotator
        annotator_named_scores = collections.defaultdict(lambda: scores_type())
        for i in range(len(annotator_data)):
            annotator1, data1 = annotator_data[i]
            for j in range(i + 1, len(annotator_data)):
                annotator2, data2 = annotator_data[j]

                # make a prefix for this specific pair of annotators
                prefix = make_prefix([annotator1, annotator2])

                # make a prefix where non-gold annotators are just called "annotator"
                general_prefix = make_prefix(
                    a if a == "gold" else "annotator" for a in [annotator1, annotator2])

                # perform the comparison of the two annotation sets and update the overall scores
                named_scores = score_data(data1, data2, include, exclude,
                                          scores_type=scores_type, spans_type=spans_type)

                # add annotators as prefixes
                for name, scores in named_scores.items():
                    if not isinstance(name, tuple):
                        name = name,
                    annotator_named_scores[(prefix,) + name].update(scores)
                    annotator_named_scores[(general_prefix,) + name].update(scores)

        # generate the filename and the resulting scores
        yield text_name, annotator_named_scores


def _print_document_scores(file_named_scores):

    def _score_name(x):
        return ":".join(x) if isinstance(x, tuple) else x

    print("{0:40}\t{1:40}\t{2:^5}\t{3:^5}\t{4:^5}\t{5:^5}\t{6:^5}\t{7:^5}".format(
        "", "", "ref", "pred", "corr", "P", "R", "F1"))
    for file_name, named_scores in file_named_scores:
        for name, scores in named_scores.items():
            print("{0!s:40}\t{1!s:40}\t{2!s:5}\t{3!s:5}\t{4!s:5}\t{5:5.3f}\t{6:5.3f}\t{7:5.3f}".format(
                file_name, _score_name(name), scores.reference, scores.predicted, scores.correct,
                scores.precision(), scores.recall(), scores.f1()))


def _print_merged_scores(file_named_scores, scores_type):
    all_named_scores = collections.defaultdict(lambda: scores_type())
    for _, named_scores in file_named_scores:
        for name, scores in named_scores.items():
            all_named_scores[name].update(scores)

    def _score_name(x):
        return ":".join(x) if isinstance(x, tuple) else x

    print("{0:40}\t{1:^5}\t{2:^5}\t{3:^5}\t{4:^5}\t{5:^5}\t{6:^5}".format(
        "", "ref", "pred", "corr", "P", "R", "F1"))
    for name in sorted(all_named_scores, key=_score_name):
        scores = all_named_scores[name]
        print("{0!s:40}\t{1!s:5}\t{2!s:5}\t{3!s:5}\t{4:5.3f}\t{5:5.3f}\t{6:5.3f}".format(
            _score_name(name), scores.reference, scores.predicted, scores.correct,
            scores.precision(), scores.recall(), scores.f1()))


if __name__ == "__main__":
    def split_tuple_on_colons(string):
        result = tuple(string.split(":"))
        return result[0] if len(result) == 1 else result

    parser = argparse.ArgumentParser(description="""%(prog)s compares one directory of Anafora XML annotations to
        another and prints statistics such as precision, recall and F-measure. It can also be used with a single
        Anafora XML directory to compute inter-annotator agreement.""")
    parser.set_defaults(scores_type=Scores)
    parser.add_argument("-r", "--reference", metavar="DIR", dest="reference_dir", required=True,
                        help="The root of a set of Anafora XML directories representing reference annotations.")
    parser.add_argument("-p", "--predicted", metavar="DIR", dest="predicted_dir",
                        help="The root of a set of Anafora XML directories representing system-predicted annotations.")
    parser.add_argument("-t", "--text", metavar="DIR", dest="text_dir",
                        help="A flat directory containing the raw text. By default, the reference directory is " +
                             "assumed to contain the raw text. (Text is typically only needed with --verbose.)")
    parser.add_argument("-i", "--include", metavar="EXPR", nargs="+", type=split_tuple_on_colons,
                        help="An expression identifying types of annotations to be included in the evaluation. " +
                             "The expression takes the form type[:property[:value]. For example, TLINK would only " +
                             "include TLINK annotations (and TLINK properties and property values) in the " +
                             "evaluation, while TLINK:Type:CONTAINS would only include TLINK annotations with a Type " +
                             "property that has the value CONTAINS.")
    parser.add_argument("-e", "--exclude", metavar="EXPR", nargs="+", type=split_tuple_on_colons,
                        help="An expression identifying types of annotations to be excluded from the evaluation. " +
                             "The expression takes the form type[:property[:value] (see --include).")
    parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                        help="A regular expression for matching XML files in the subdirectories, typically used to " +
                             "restrict the evaluation to a subset of the available files (default: %(default)r)")
    parser.add_argument("--temporal-closure", action="store_const", const=TemporalClosureScores, dest="scores_type",
                        help="Apply temporal closure on the reference annotations when calculating precision, and " +
                             "apply temporal closure on the predicted annotations when calculating recall. " +
                             "This must be combined with --include to restrict the evaluation to a Type:Property " +
                             "whose values are valid temporal relations (BEFORE, AFTER, INCLUDES, etc.)")
    parser.add_argument("--per-document", action="store_true",
                        help="Print out scores for each document, rather than overall scores")
    parser.add_argument("--verbose", action="store_const", const=DebuggingScores, dest="scores_type",
                        help="Include more information in the output, such as the reference expressions that were " +
                             "and the predicted expressions that were not in the reference.")
    parser.add_argument("--overlap", dest="spans_type", action="store_const", const=_OverlappingSpans,
                        help="Count predicted annotation spans as correct if they overlap by one character or more " +
                             "with a reference annotation span. Not intended as a real evaluation method (since what " +
                             "to do with multiple matches is not well defined) but useful for debugging purposes.")
    args = parser.parse_args()
    basic_config_kwargs = {"format": "%(levelname)s:%(message)s"}
    if args.scores_type == DebuggingScores:
        basic_config_kwargs["level"] = logging.DEBUG
    logging.basicConfig(**basic_config_kwargs)

    if args.predicted_dir is not None:
        _file_named_scores = score_dirs(
            reference_dir=args.reference_dir,
            predicted_dir=args.predicted_dir,
            xml_name_regex=args.xml_name_regex,
            text_dir=args.text_dir,
            include=args.include,
            exclude=args.exclude,
            scores_type=args.scores_type,
            spans_type=args.spans_type)
    else:
        _file_named_scores = score_annotators(
            anafora_dir=args.reference_dir,
            xml_name_regex=args.xml_name_regex,
            include=args.include,
            exclude=args.exclude,
            scores_type=args.scores_type,
            spans_type=args.spans_type)

    if args.per_document:
        _print_document_scores(_file_named_scores)
    else:
        _print_merged_scores(_file_named_scores, scores_type=args.scores_type)



"""Logic that powers autocompletion installed by ``pip completion``.
"""

import optparse
import os
import sys
from itertools import chain
from typing import Any, Iterable, List, Optional

from pip._internal.cli.main_parser import create_main_parser
from pip._internal.commands import commands_dict, create_command
from pip._internal.metadata import get_default_environment


def autocomplete() -> None:
    """Entry Point for completion of main and subcommand options."""
    # Don't complete if user hasn't sourced bash_completion file.
    if "PIP_AUTO_COMPLETE" not in os.environ:
        return
    cwords = os.environ["COMP_WORDS"].split()[1:]
    cword = int(os.environ["COMP_CWORD"])
    try:
        current = cwords[cword - 1]
    except IndexError:
        current = ""

    parser = create_main_parser()
    subcommands = list(commands_dict)
    options = []

    # subcommand
    subcommand_name: Optional[str] = None
    for word in cwords:
        if word in subcommands:
            subcommand_name = word
            break
    # subcommand options
    if subcommand_name is not None:
        # special case: 'help' subcommand has no options
        if subcommand_name == "help":
            sys.exit(1)
        # special case: list locally installed dists for show and uninstall
        should_list_installed = not current.startswith("-") and subcommand_name in [
            "show",
            "uninstall",
        ]
        if should_list_installed:
            env = get_default_environment()
            lc = current.lower()
            installed = [
                dist.canonical_name
                for dist in env.iter_installed_distributions(local_only=True)
                if dist.canonical_name.startswith(lc)
                and dist.canonical_name not in cwords[1:]
            ]
            # if there are no dists installed, fall back to option completion
            if installed:
                for dist in installed:
                    print(dist)
                sys.exit(1)

        subcommand = create_command(subcommand_name)

        for opt in subcommand.parser.option_list_all:
            if opt.help != optparse.SUPPRESS_HELP:
                for opt_str in opt._long_opts + opt._short_opts:
                    options.append((opt_str, opt.nargs))

        # filter out previously specified options from available options
        prev_opts = [x.split("=")[0] for x in cwords[1 : cword - 1]]
        options = [(x, v) for (x, v) in options if x not in prev_opts]
        # filter options by current input
        options = [(k, v) for k, v in options if k.startswith(current)]
        # get completion type given cwords and available subcommand options
        completion_type = get_path_completion_type(
            cwords,
            cword,
            subcommand.parser.option_list_all,
        )
        # get completion files and directories if ``completion_type`` is
        # ``<file>``, ``<dir>`` or ``<path>``
        if completion_type:
            paths = auto_complete_paths(current, completion_type)
            options = [(path, 0) for path in paths]
        for option in options:
            opt_label = option[0]
            # append '=' to options which require args
            if option[1] and option[0][:2] == "--":
                opt_label += "="
            print(opt_label)
    else:
        # show main parser options only when necessary

        opts = [i.option_list for i in parser.option_groups]
        opts.append(parser.option_list)
        flattened_opts = chain.from_iterable(opts)
        if current.startswith("-"):
            for opt in flattened_opts:
                if opt.help != optparse.SUPPRESS_HELP:
                    subcommands += opt._long_opts + opt._short_opts
        else:
            # get completion type given cwords and all available options
            completion_type = get_path_completion_type(cwords, cword, flattened_opts)
            if completion_type:
                subcommands = list(auto_complete_paths(current, completion_type))

        print(" ".join([x for x in subcommands if x.startswith(current)]))
    sys.exit(1)


def get_path_completion_type(
    cwords: List[str], cword: int, opts: Iterable[Any]
) -> Optional[str]:
    """Get the type of path completion (``file``, ``dir``, ``path`` or None)

    :param cwords: same as the environmental variable ``COMP_WORDS``
    :param cword: same as the environmental variable ``COMP_CWORD``
    :param opts: The available options to check
    :return: path completion type (``file``, ``dir``, ``path`` or None)
    """
    if cword < 2 or not cwords[cword - 2].startswith("-"):
        return None
    for opt in opts:
        if opt.help == optparse.SUPPRESS_HELP:
            continue
        for o in str(opt).split("/"):
            if cwords[cword - 2].split("=")[0] == o:
                if not opt.metavar or any(
                    x in ("path", "file", "dir") for x in opt.metavar.split("/")
                ):
                    return opt.metavar
    return None


def auto_complete_paths(current: str, completion_type: str) -> Iterable[str]:
    """If ``completion_type`` is ``file`` or ``path``, list all regular files
    and directories starting with ``current``; otherwise only list directories
    starting with ``current``.

    :param current: The word to be completed
    :param completion_type: path completion type(`file`, `path` or `dir`)i
    :return: A generator of regular files and/or directories
    """
    directory, filename = os.path.split(current)
    current_path = os.path.abspath(directory)
    # Don't complete paths if they can't be accessed
    if not os.access(current_path, os.R_OK):
        return
    filename = os.path.normcase(filename)
    # list all files that start with ``filename``
    file_list = (
        x for x in os.listdir(current_path) if os.path.normcase(x).startswith(filename)
    )
    for f in file_list:
        opt = os.path.join(current_path, f)
        comp_file = os.path.normcase(os.path.join(directory, f))
        # complete regular files when there is not ``<dir>`` after option
        # complete directories when there is ``<file>``, ``<path>`` or
        # ``<dir>``after option
        if completion_type != "dir" and os.path.isfile(opt):
            yield comp_file
        elif os.path.isdir(opt):
            yield os.path.join(comp_file, "")



from __future__ import absolute_import

import argparse
import logging
import os
import regex

import anafora
import anafora.validate

def fix_thyme_errors(schema, input_dir, output_dir, xml_name_regex="[.]xml$"):
    """
    :param schema anafora.validate.Schema: the THYME schema
    :param input_dir str: the root of a set of THYME Anafora XML directories
    :param output_dir str: the directory where the cleaned versions of the THYME Anafora XML files should be written.
        The directory structure will mirror the input directory structure.
    """
    for sub_dir, text_name, xml_names in anafora.walk(input_dir, xml_name_regex):
        for xml_name in xml_names:
            xml_path = os.path.join(input_dir, sub_dir, xml_name)

            # load the data from the Anafora XML
            try:
                data = anafora.AnaforaData.from_file(xml_path)
            except anafora.ElementTree.ParseError as e:
                logging.warning("SKIPPING invalid XML: %s: %s", e, xml_path)
                continue

            # remove invalid TLINKs and ALINKs
            changed = False
            to_remove = []
            for annotation in data.annotations:
                try:
                    schema.validate(annotation)
                except anafora.validate.SchemaValidationError as e:
                    if annotation.type in {"TLINK", "ALINK"}:
                        logging.warning("REMOVING %s: %s", e, annotation)
                        to_remove.append(annotation)
            for annotation in to_remove:
                data.annotations.remove(annotation)
                changed = True

            # remove TIMEX3s that are directly on top of SECTIONTIMEs and DOCTIMEs
            for span, annotations in anafora.validate.find_entities_with_identical_spans(data):
                try:
                    # sorts SECTIONTIME and DOCTIME before TIMEX3
                    special_time, timex = sorted(annotations, key=lambda a: a.type)
                except ValueError:
                    pass
                else:
                    if special_time.type in {"SECTIONTIME", "DOCTIME"} and timex.type == "TIMEX3":
                        msg = "REPLACING multiple entities for span %s: %s WITH %s"
                        logging.warning(msg, span, timex, special_time)
                        for annotation in data.annotations:
                            for name, value in annotation.properties.items():
                                if value is timex:
                                    annotation.properties[name] = special_time
                        data.annotations.remove(timex)
                        changed = True

            # if we found and fixed any errors, write out the new XML file
            if changed:
                output_sub_dir = os.path.join(output_dir, sub_dir)
                if not os.path.exists(output_sub_dir):
                    os.makedirs(output_sub_dir)
                output_path = os.path.join(output_sub_dir, xml_name)
                data.to_file(output_path)


def convert_thyme_qa_to_anafora_xml(input_dir, output_dir):
    _header_sep_pattern = regex.compile(r'\s*=====+\s*')
    _annotation_sep_pattern = regex.compile(r'\s*-----+\s*')
    _annotation_pattern = regex.compile(r'^Question:(.*?)\nAnswer:(.*?)\nConfidence:(.*?)\n' +
                                        r'Difficulty:(.*?)\nDocTimeRel:(.*?)\n(Text Clip:.*)$', regex.DOTALL)
    _text_clip_pattern = regex.compile(r'Text Clip:\s+\d[\w.]*\s+(\d+),(\d+) (Exact|Support)_Answer ' +
                                       r'Use_(Time_Span|DocTimeRel) ?(.*)\n(.*)(?:\n|$)')

    # iterate through all _qa.txt files in the input directory
    for input_root, dir_names, input_file_names in os.walk(input_dir):
        for input_file_name in input_file_names:
            if input_file_name.endswith("_qa.txt"):
                file_base = input_file_name[:-7]

                # create one Anafora XML for each file
                data = anafora.AnaforaData()
                relation_count = 1
                entity_count = 1
                with open(os.path.join(input_root, input_file_name)) as input_file:
                    text = input_file.read().decode('ascii')

                    # parse the annotations from the THYME question-answer format
                    _, body_text = _header_sep_pattern.split(text)
                    for annotation_text in _annotation_sep_pattern.split(body_text.rstrip(" \n\r-")):
                        match = _annotation_pattern.match(annotation_text)
                        if match is None:
                            raise ValueError("Invalid annotation text:\n" + annotation_text)
                        groups = [s.strip() for s in match.groups()]
                        question, answer, confidence, difficulty, doc_time_rel, text_clip_text = groups
                        text_clip_matches = _text_clip_pattern.findall(text_clip_text)
                        if len(text_clip_text.splitlines()) != 2 * len(text_clip_matches):
                            raise ValueError("Invalid Text Clips in annotation text:\n" + annotation_text)

                        # create Anafora XML annotations for the answers
                        entities = []
                        for begin_text, end_text, _, time_or_doc_time_rel, type_text, clip_text in text_clip_matches:
                            begin = int(begin_text)
                            end = int(end_text)
                            entity_annotation = anafora.AnaforaEntity()
                            entity_annotation.id = '{0:d}@{1}@{2}@gold'.format(entity_count, 'e', file_base)
                            entity_annotation.spans = ((begin, end),)
                            entity_annotation.type = 'EVENT'
                            entity_annotation.parents_type = 'TemporalEntities'
                            if time_or_doc_time_rel == 'DocTimeRel':
                                entity_annotation.properties['DocTimeRel'] = doc_time_rel.upper()
                            entity_count += 1
                            data.annotations.append(entity_annotation)
                            entities.append(entity_annotation)

                        # create an Anafora XML annotation for the question
                        question_annotation = anafora.AnaforaRelation()
                        question_annotation.id = '{0:d}@{1}@{2}@gold'.format(relation_count, 'r', file_base)
                        question_annotation.type = 'Question'
                        question_annotation.parents_type = 'TemporalQuestions'
                        question_annotation.properties['Question'] = question
                        question_annotation.properties['Confidence'] = confidence
                        question_annotation.properties['Difficulty'] = difficulty
                        # FIXME: hacking XML here because current API doesn't allow properties with multiple values
                        for entity in entities:
                            property_elem = anafora.ElementTree.SubElement(question_annotation.properties.xml, 'Answer')
                            property_elem.text = entity.id
                        data.annotations.append(question_annotation)
                        relation_count += 1

                # write the Anafora data out as XML
                output_file_dir = os.path.join(output_dir, file_base)
                output_file_path = os.path.join(output_file_dir, file_base + ".THYME_QA.preannotation.completed.xml")
                if not os.path.exists(output_file_dir):
                    os.makedirs(output_file_dir)
                data.indent()
                data.to_file(output_file_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    thyme_parser = subparsers.add_parser('thyme')
    thyme_parser.set_defaults(func=fix_thyme_errors)
    thyme_parser.add_argument("-s", "--schema", metavar="FILE", required=True, type=anafora.validate.Schema.from_file,
                              help="An Anafora schema XML file which Anafora annotation XML files will be validated " +
                                   "against.")
    thyme_parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="input_dir",
                              help="The root of a set of Anafora annotation XML directories.")
    thyme_parser.add_argument("-o", "--output", metavar="DIR", required=True, dest="output_dir",
                              help="The directory where the cleaned versions of the Anafora annotation XML files " +
                                   "should be written. The directory structure will mirror the input directory " +
                                   "structure.")
    thyme_parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                              help="A regular expression for matching XML files, typically used to restrict the " +
                                   "validation to a subset of the available files (default: %(default)r)")

    thyme_qa_parser = subparsers.add_parser('thyme-qa')
    thyme_qa_parser.set_defaults(func=convert_thyme_qa_to_anafora_xml)
    thyme_qa_parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="input_dir",
                                 help="The root of a set of THYME _qa.txt files.")
    thyme_qa_parser.add_argument("-o", "--output", metavar="DIR", required=True, dest="output_dir",
                                 help="The directory where the Anafora annotation XML files should be written.")

    args = parser.parse_args()
    kwargs = vars(args)
    kwargs.pop('func')(**kwargs)



import argparse
import os
import re

import sys

import anafora
import anafora.timeml


def copy_timeml_text(text_dir, anafora_dir, xml_name_regex, write_dct):
    text_name_to_path = {}
    text_name_to_folder = {}
    for dir_path, _, file_names in os.walk(text_dir):
        for file_name in file_names:
            if file_name.endswith(".tml"):
                file_path = os.path.join(dir_path, file_name)
                text_name_to_path[file_name[:-4]] = file_path
                file_folder = os.path.relpath(dir_path, text_dir)
                file_folder = os.path.join(file_folder, file_name[:-4])
                text_name_to_folder[file_name[:-4]] = file_folder

    def get_dct(path):
        dct = anafora.timeml.to_document_creation_time(path)
        return dct[:10]

    _copy_text(text_name_to_path=text_name_to_path,
               text_name_to_folder=text_name_to_folder,
               get_text=anafora.timeml.to_text,
               get_dct=get_dct if write_dct else None,
               anafora_dir=anafora_dir,
               xml_name_regex=xml_name_regex)


def copy_plain_text(text_dir, anafora_dir, xml_name_regex, write_dct):
    if write_dct:
        raise ValueError("write_dct is not supported for plain text")
    text_name_to_path = {}
    text_name_to_folder = {}
    for dir_path, _, file_names in os.walk(text_dir):
        for file_name in file_names:
            text_name_to_path[file_name] = os.path.join(dir_path, file_name)
            file_folder = os.path.relpath(dir_path, text_dir)
            file_folder = os.path.join(file_folder, file_name)
            text_name_to_folder[file_name] = file_folder
    _copy_text(text_name_to_path=text_name_to_path,
               text_name_to_folder=text_name_to_folder,
               get_text=lambda path: open(path).read(),
               get_dct=None,
               anafora_dir=anafora_dir,
               xml_name_regex=xml_name_regex)


def copy_mayo_text(text_dir, anafora_dir, xml_name_regex, write_dct):
    text_name_to_path = {}
    text_name_to_folder = {}
    for dir_path, _, file_names in os.walk(text_dir):
        for file_name in file_names:
            text_name_to_path[file_name] = os.path.join(dir_path, file_name)
            file_folder = os.path.relpath(dir_path, text_dir)
            file_folder = os.path.join(file_folder, file_name)
            text_name_to_folder[file_name] = file_folder

    def get_dct(path):
        dct_pattern = r"\[meta rev_date=\"(?P<m>\d+)/(?P<d>\d+)/(?P<y>\d+)\""
        with open(path) as f:
            match = re.match(dct_pattern, f.read())
        kwargs = {name: int(value) for name, value in match.groupdict().items()}
        return "{y}-{m:02}-{d:02}".format(**kwargs)

    _copy_text(text_name_to_path=text_name_to_path,
               text_name_to_folder=text_name_to_folder,
               get_text=lambda path: open(path).read(),
               get_dct=get_dct if write_dct else None,
               anafora_dir=anafora_dir,
               xml_name_regex=xml_name_regex)


def _copy_text(text_name_to_path,
               text_name_to_folder,
               get_text,
               get_dct,
               anafora_dir,
               xml_name_regex):
    if not os.listdir(anafora_dir):
        _copy_text_output_dir(text_name_to_path,
                              get_text,
                              get_dct,
                              anafora_dir,
                              text_name_to_folder)
    else:
        _copy_text_anafora_dir(text_name_to_path,
                               get_text,
                               get_dct,
                               anafora_dir,
                               xml_name_regex)


def _copy_text_anafora_dir(text_name_to_path,
                           get_text,
                           get_dct,
                           anafora_dir,
                           xml_name_regex):
    for sub_dir, text_file_name, _ in anafora.walk(
            anafora_dir, xml_name_regex=xml_name_regex):
        if text_file_name not in text_name_to_path:
            sys.exit("No text file found for " + text_file_name)
        text_path = os.path.join(anafora_dir, sub_dir, text_file_name)
        if os.path.exists(text_path):
            sys.exit("Text file already exists: " + text_path)
        input_path = text_name_to_path[text_file_name]
        text = get_text(input_path)
        with open(text_path, 'w') as text_file:
            text_file.write(text)
        if get_dct is not None:
            _copy_dct(get_dct, input_path, text_path)


def _copy_text_output_dir(text_name_to_path,
                          get_text,
                          get_dct,
                          output_dir,
                          text_name_to_folder):
    for text_file_name in text_name_to_path:
        text_file_folder = text_name_to_folder[text_file_name]
        output_path = os.path.join(output_dir, text_file_folder)
        text_path = os.path.join(output_path, text_file_name)
        if os.path.exists(text_path):
            sys.exit("Text file already exists: " + text_path)
        input_path = text_name_to_path[text_file_name]
        text = get_text(input_path)
        os.makedirs(output_path, exist_ok=True)
        with open(text_path, 'w') as text_file:
            text_file.write(text)
        if get_dct is not None:
            _copy_dct(get_dct, input_path, text_path)


def _copy_dct(get_dct, input_path, text_path):
    dct_path = text_path + ".dct"
    if os.path.exists(dct_path):
        sys.exit("DCT file already exists: " + dct_path)
    dct = get_dct(input_path)
    with open(dct_path, 'w') as dct_file:
        dct_file.write(dct)
        dct_file.write("\n")


if __name__ == "__main__":
    format_funcs = dict(plain=copy_plain_text,
                        mayo=copy_mayo_text,
                        timeml=copy_timeml_text)

    parser = argparse.ArgumentParser(
        description="Copies text files to their expected locations in an " +
                    "Anafora XML directory hierarchy. That is, each text " +
                    "file that exactly matches the name of a directory in " +
                    "the Anafora XML directory hierarchy will be copied into " +
                    "that directory. If Anafora XML directory is emtpy, all " +
                    "the text files will be copied.")
    parser.add_argument(
        "-x", "--xml-name-regex",
        metavar="REGEX", default=".*completed.*[.]xml$",
        help="The regular expression which identifies Anafora XML files " +
             "(default: %(default)r).")
    parser.add_argument(
        "--format", choices=format_funcs, default="plain",
        help="The format of files in the text directory " +
             "(default: %(default)r). When the 'timeml' format is selected, " +
             "TimeML files are expected to have the extension '.tml', and " +
             "the text of a '.tml' file is the result of stripping that file " +
             "of all its XML tags. The 'mayo' and 'plain' formats both " +
             "simply copy over the text files directly; the only difference " +
             "is that the 'mayo' format also supports the --dct flag.")
    parser.add_argument(
        "--dct", action="store_true", dest="write_dct",
        help="Also write out a file for the document creation time.")
    parser.add_argument("text_dir")
    parser.add_argument("anafora_dir")

    kwargs = vars(parser.parse_args())
    format_funcs[kwargs.pop("format")](**kwargs)



import argparse
import os

import anafora


def _flatten_to_ints(items):
    for item in items:
        if not isinstance(item, int):
            for sub_item in _flatten_to_ints(item):
                yield sub_item
        else:
            yield item


def add_relations_to_closest(data, source_type, target_type,
                             relation_type, relation_source_property_name, relation_target_property_name,
                             relation_other_properties=None):
    """
    Adds a relation from each `source_type` annotation to the closest `target_type` annotation.

    :param anafora.AnaforaData data: the Anafora data where relations should be added
    :param str source_type: the type of the source annotations
    :param str target_type: the type of the target annotations
    :param str relation_type: the type of relation annotation to be created
    :param str relation_source_property_name: the name of the property on the relation annotation that should point to
        the source annotation
    :param str relation_target_property_name: the name of the property on the relation annotation that should point to
        the target annotation
    :param list relation_other_properties: a list of (name, value) tuples of other properties that should be set on the
        relation annotations that are created
    """

    # map the id of each source/target annotation to its character offsets
    points = {}
    for source_entity in data.annotations.select_type(source_type):
        points[source_entity.id] = list(_flatten_to_ints(source_entity.spans))
    for target_entity in data.annotations.select_type(target_type):
        points[target_entity.id] = list(_flatten_to_ints(target_entity.spans))

    # add a relation for each source entity
    target_entities = list(data.annotations.select_type(target_type))
    if target_entities:
        source_entities = list(data.annotations.select_type(source_type))
        for source_entity in source_entities:

            # distance to an annotation is the minimum distance to any one of its character offsets
            def distance_to_source_entity(entity):
                return min(abs(p1 - p2) for p1 in points[source_entity.id] for p2 in points[entity.id])

            # find the target entity that is closest to the source entity
            target_entity = min(target_entities, key=distance_to_source_entity)

            # create a relation annotation per the various arguments to this function
            relation = anafora.AnaforaRelation()
            relation.id = "{0}@{1}@{2}".format(source_entity.id, relation_type, target_entity.id)
            data.annotations.append(relation)
            relation.type = relation_type
            relation.properties[relation_source_property_name] = source_entity
            relation.properties[relation_target_property_name] = target_entity
            if relation_other_properties is not None:
                for name, value in relation_other_properties:
                    relation.properties[name] = value

    data.indent()


if __name__ == "__main__":
    def _pair(value):
        return value.split("=")

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="input_dir",
                        help="The root of a set of Anafora annotation XML directories.")
    parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                        help="A regular expression for matching XML files, typically used to restrict the validation " +
                             "to a subset of the available files (default: %(default)r)")
    parser.add_argument("-o", "--output", metavar="DIR", required=True, dest="output_dir",
                        help="The directory where the predictions should be written (as Anafora XML files). The " +
                             "directory structure will mirror the input directory structure.")

    subparsers = parser.add_subparsers()

    relations_to_closest_parser = subparsers.add_parser('closest', help="Link each source annotation to the target " +
                                                                        "annotation with the closest character offsets")
    relations_to_closest_parser.set_defaults(func=add_relations_to_closest)
    relations_to_closest_parser.add_argument("-s", "--source", metavar="TYPE", dest="source_type", required=True,
                                             help="The <type> of the source annotations.")
    relations_to_closest_parser.add_argument("-t", "--target", metavar="TYPE", dest="target_type", required=True,
                                             help="The <type> of the target annotations.")
    relations_to_closest_parser.add_argument("-r", "--relation", metavar="TYPE", dest="relation_type", required=True,
                                             help="The <type> of relation annotation to be created.")
    relations_to_closest_parser.add_argument("-rs", "--relation-source", metavar="NAME", required=True,
                                             dest="relation_source_property_name",
                                             help="The name of the property on the relation annotation that should " +
                                                  "point to the source annotation.")
    relations_to_closest_parser.add_argument("-rt", "--relation-target", metavar="NAME", required=True,
                                             dest="relation_target_property_name",
                                             help="The name of the property on the relation annotation that should " +
                                                  "point to the target annotation.")
    relations_to_closest_parser.add_argument("-ro", "--relation-other", metavar="NAME=VALUE", nargs='+', type=_pair,
                                             dest="relation_other_properties",
                                             help="Other properties that should be added to the relation annotation.")

    args = parser.parse_args()
    kwargs = vars(args)
    func = kwargs.pop("func")
    input_dir = kwargs.pop('input_dir')
    xml_name_regex = kwargs.pop('xml_name_regex')
    output_dir = kwargs.pop('output_dir')

    for sub_dir, _, xml_file_names in anafora.walk(input_dir, xml_name_regex):
        for xml_file_name in xml_file_names:
            input_data = anafora.AnaforaData.from_file(os.path.join(input_dir, sub_dir, xml_file_name))
            func(input_data, **kwargs)
            output_sub_dir = os.path.join(output_dir, sub_dir)
            if not os.path.exists(output_sub_dir):
                os.makedirs(output_sub_dir)
            input_data.to_file(os.path.join(output_dir, sub_dir, xml_file_name))
